# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for releasing values from a federated program to memory."""

import collections
from collections.abc import Hashable

from federated_language.program import release_manager
from federated_language.program import value_reference


class MemoryReleaseManager(
    release_manager.ReleaseManager[
        release_manager.ReleasableStructure, Hashable
    ]
):
  """A `federated_language.program.ReleaseManager` that releases values to memory.

  A `federated_language.program.MemoryReleaseManager` is a utility for releasing
  values from a
  federated program to memory and is used to release values from platform
  storage to customer storage in a federated program.

  Values are released to memory as Python objects. When the value is released,
  if the value is a value reference or a structure containing value references,
  each value reference is materialized.
  """

  def __init__(self):
    """Returns an initialized `federated_language.program.MemoryReleaseManager`."""
    self._values = collections.OrderedDict()

  async def release(
      self, value: release_manager.ReleasableStructure, key: Hashable
  ) -> None:
    """Releases `value` from a federated program.

    Args:
      value: A `federated_language.program.ReleasableStructure` to release.
      key: A hashable value used to reference the released `value`.
    """
    materialized_value = await value_reference.materialize_value(value)
    self._values[key] = materialized_value

  def remove_all(self) -> None:
    """Removes all program states."""
    self._values = collections.OrderedDict()

  def values(
      self,
  ) -> collections.OrderedDict[Hashable, release_manager.ReleasableStructure]:
    """Returns an `collections.OrderedDict` of all keys and released values."""
    return self._values.copy()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A context for execution based on an embedded executor instance."""

from collections.abc import Callable
from typing import Optional

from federated_language.common_libs import async_utils
from federated_language.computation import computation_base
from federated_language.context_stack import context
from federated_language.executor import async_execution_context
from federated_language.executor import cardinalities_utils
from federated_language.executor import executor_factory  # pylint: disable=unused-import


class SyncExecutionContext(context.SyncContext):
  """A synchronous execution context backed by an `executor_base.Executor`."""

  def __init__(
      self,
      executor_fn: executor_factory.ExecutorFactory,
      compiler_fn: Optional[
          Callable[[computation_base.Computation], object]
      ] = None,
      *,
      transform_args: Optional[Callable[[object], object]] = None,
      transform_result: Optional[Callable[[object], object]] = None,
      cardinality_inference_fn: cardinalities_utils.CardinalityInferenceFnType = cardinalities_utils.infer_cardinalities,
  ):
    """Initializes a synchronous execution context which retries invocations.

    Args:
      executor_fn: Instance of `executor_factory.ExecutorFactory`.
      compiler_fn: A Python function that will be used to compile a computation.
      transform_args: An `Optional` `Callable` used to transform the args before
        they are passed to the computation.
      transform_result: An `Optional` `Callable` used to transform the result
        before it is returned.
      cardinality_inference_fn: A Python function specifying how to infer
        cardinalities from arguments (and their associated types). The value
        returned by this function will be passed to the `create_executor` method
        of `executor_fn` to construct a `federated_language.framework.Executor`
        instance.
    """
    self._executor_factory = executor_fn
    self._async_context = async_execution_context.AsyncExecutionContext(
        executor_fn=executor_fn,
        compiler_fn=compiler_fn,
        transform_args=transform_args,
        transform_result=transform_result,
        cardinality_inference_fn=cardinality_inference_fn,
    )
    self._async_runner = async_utils.AsyncThreadRunner()

  @property
  def executor_factory(self) -> executor_factory.ExecutorFactory:
    return self._executor_factory

  def invoke(self, comp, arg):
    return self._async_runner.run_coro_and_return_result(
        self._async_context.invoke(comp, arg)
    )



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Libraries for creating federated programs."""

# pylint: disable=g-importing-member
from federated_language.program.data_source import FederatedDataSource
from federated_language.program.data_source import FederatedDataSourceIterator
from federated_language.program.federated_context import check_in_federated_context
from federated_language.program.federated_context import ComputationArg
from federated_language.program.federated_context import contains_only_server_placed_data
from federated_language.program.federated_context import FederatedContext
from federated_language.program.logging_release_manager import LoggingReleaseManager
from federated_language.program.memory_release_manager import MemoryReleaseManager
from federated_language.program.native_platform import NativeFederatedContext
from federated_language.program.native_platform import NativeValueReference
from federated_language.program.program_state_manager import ProgramStateExistsError
from federated_language.program.program_state_manager import ProgramStateManager
from federated_language.program.program_state_manager import ProgramStateNotFoundError
from federated_language.program.program_state_manager import ProgramStateStructure
from federated_language.program.program_state_manager import ProgramStateValue
from federated_language.program.release_manager import DelayedReleaseManager
from federated_language.program.release_manager import FilteringReleaseManager
from federated_language.program.release_manager import GroupingReleaseManager
from federated_language.program.release_manager import Key
from federated_language.program.release_manager import NotFilterableError
from federated_language.program.release_manager import PeriodicReleaseManager
from federated_language.program.release_manager import ReleasableStructure
from federated_language.program.release_manager import ReleasableValue
from federated_language.program.release_manager import ReleasedValueNotFoundError
from federated_language.program.release_manager import ReleaseManager
from federated_language.program.value_reference import MaterializableStructure
from federated_language.program.value_reference import MaterializableTypeSignature
from federated_language.program.value_reference import MaterializableValue
from federated_language.program.value_reference import MaterializableValueReference
from federated_language.program.value_reference import materialize_value
from federated_language.program.value_reference import MaterializedStructure
from federated_language.program.value_reference import MaterializedValue
# pylint: enable=g-importing-member



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities file for functions with `Value`s as inputs and outputs."""

from typing import Optional

from federated_language.compiler import building_block_factory
from federated_language.compiler import building_blocks
from federated_language.federated_context import value_impl
from federated_language.types import computation_types
from federated_language.types import placements


def get_curried(fn: value_impl.Value):
  """Returns a curried version of function `fn` that takes a parameter tuple.

  For functions `fn` of types <T1,T2,....,Tn> -> U, the result is a function
  of the form T1 -> (T2 -> (T3 -> .... (Tn -> U) ... )).

  NOTE: No attempt is made at avoiding naming conflicts in cases where `fn`
  contains references. The arguments of the curriend function are named `argN`
  with `N` starting at 0.

  Args:
    fn: A value of a functional type.

  Returns:
    A value that represents the curried form of `fn`.
  """
  if not isinstance(fn.type_signature, computation_types.FunctionType):
    raise ValueError(
        'Expected a `federated_language.FunctionType`, found'
        f' {fn.type_signature}.'
    )
  if not isinstance(fn.type_signature.parameter, computation_types.StructType):
    raise ValueError(
        'Expected a `federated_language.StructType`, found'
        f' {fn.type_signature.parameter}.'
    )
  param_elements = fn.type_signature.parameter.items()
  references = []
  for idx, (_, elem_type) in enumerate(param_elements):
    references.append(building_blocks.Reference('arg{}'.format(idx), elem_type))
  result = building_blocks.Call(fn.comp, building_blocks.Struct(references))
  for ref in references[::-1]:
    result = building_blocks.Lambda(ref.name, ref.type_signature, result)
  return value_impl.Value(result)


def ensure_federated_value(
    value: value_impl.Value,
    placement: Optional[placements.PlacementLiteral] = None,
    label: Optional[str] = None,
):
  """Ensures `value` is a federated value placed at `placement`.

  If `value` is not a `computation_types.FederatedType` but is a
  `computation_types.StructType` that can be converted via `federated_zip`
  to a `computation_types.FederatedType`, inserts the call to `federated_zip`
  and returns the result. If `value` cannot be converted, raises a TypeError.

  Args:
    value: A `value_impl.Value` to check and convert to a federated value if
      possible.
    placement: The expected placement. If None, any placement is allowed.
    label: An optional string label that describes `value`.

  Returns:
    The value as a federated value, automatically zipping if necessary.

  Raises:
    TypeError: if `value` is not a `FederatedType` and cannot be converted to
      a `FederatedType` with `federated_zip`.
  """

  if not isinstance(value.type_signature, computation_types.FederatedType):
    comp = value.comp
    try:
      zipped = building_block_factory.create_federated_zip(comp)
    except (TypeError, ValueError) as e:
      raise TypeError(
          'The {l} must be a FederatedType or implicitly convertible '
          'to a FederatedType (got a {t}).'.format(
              l=label if label else 'value', t=comp.type_signature
          )
      ) from e
    value = value_impl.Value(zipped)

  if placement is not None and value.type_signature.placement is not placement:  # pytype: disable=attribute-error
    raise TypeError(
        'The {} should be placed at {}, but it is placed at {}.'.format(
            label if label else 'value',
            placement,
            value.type_signature.placement,  # pytype: disable=attribute-error
        )
    )

  return value



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for working with structured data."""

from collections.abc import Callable, Iterable, Mapping, Sequence
import functools
import typing
from typing import Optional, TypeVar, Union

import attrs
from federated_language.common_libs import py_typecheck
import tree


T = TypeVar('T')
# This type defines the structures supported by the `federated_language.program`
# API, meaning values of type `T` nested in structures defined by this type. For
# an example of how to use this type see
# `federated_language.program.MaterializedStructure`.
Structure = Union[
    T,
    Sequence['Structure[T]'],
    Mapping[str, 'Structure[T]'],
]


def _filter_structure(structure: Structure[object]) -> Structure[object]:
  """Returns a filtered `federated_language.program.Structure`.

  Containers that are not explicitly supported by
  `federated_language.program.Structure` are
  filtered out (by converting them to `None`) and objects are converted to
  `None`.

  Args:
    structure: A `federated_language.program.Structure`.
  """

  structure_types = []
  for arg in typing.get_args(Structure):
    origin_type = typing.get_origin(arg)
    if origin_type is not None:
      structure_types.append(origin_type)

  def _fn(structure: Structure[object]) -> Optional[object]:
    if tree.is_nested(structure) and not attrs.has(type(structure)):
      if isinstance(structure, tuple(structure_types)):
        return None
      else:
        return tree.MAP_TO_NONE
    else:
      return tree.MAP_TO_NONE

  return tree.traverse(_fn, structure)


def flatten_with_name(structure: Structure[T]) -> list[tuple[str, T]]:
  """Creates a flattened representation of the `structure` with names.

  Args:
    structure: A `federated_language.program.Structure`.

  Returns:
    A `list` of `(name, value)` `tuples` representing the flattened `structure`,
    where `name` uniquely identifies the position of the `value` in the
    `structure`.
  """
  filtered_structure = _filter_structure(structure)
  flattened = tree.flatten_with_path_up_to(filtered_structure, structure)

  def _name(path: Iterable[Union[int, str]]) -> str:
    return '/'.join(map(str, path))

  return [(_name(path), value) for path, value in flattened]


def flatten(structure: Structure[T]) -> list[T]:
  """Flattens a `federated_language.program.Structure` into a `list`."""
  filtered_structure = _filter_structure(structure)
  return tree.flatten_up_to(filtered_structure, structure)


def unflatten_as(
    structure: Structure[T], flat_sequence: Sequence[T]
) -> Structure[T]:
  """Unflattens a sequence into a `federated_language.program.Structure`."""
  filtered_structure = _filter_structure(structure)
  return tree.unflatten_as(filtered_structure, flat_sequence)


def map_structure(
    fn: Callable[..., T], *structures: Structure[T], **kwargs: object
) -> Structure[T]:
  """Maps `fn` through the `structures`."""
  if not structures:
    raise ValueError('Expected at least one structure.')
  filtered_structures = [_filter_structure(x) for x in structures]
  first_structure, *other_structures = filtered_structures
  check_types = kwargs.get('check_types', True)
  for structure in other_structures:
    tree.assert_same_structure(first_structure, structure, check_types)

  return tree.map_structure_up_to(first_structure, fn, *structures)


def map_structure_up_to(
    shallow_structure: Structure[T],
    fn: Callable[..., T],
    *structures: Structure[T],
) -> Structure[T]:
  """Maps `fn` through the `structures` up to `shallow_structure`."""
  if not structures:
    raise ValueError('Expected at least one structure.')
  filtered_structure = _filter_structure(shallow_structure)

  return tree.map_structure_up_to(filtered_structure, fn, *structures)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for Python functions, defuns, and other types of callables."""

from collections.abc import Callable
from typing import Optional

from federated_language.computation import computation_impl
from federated_language.computation import function_utils
from federated_language.types import computation_types


class PolymorphicComputation:
  """A generic polymorphic function that accepts arguments of diverse types."""

  def __init__(
      self,
      concrete_function_factory: Callable[
          [computation_types.Type, Optional[bool]],
          computation_impl.ConcreteComputation,
      ],
      infer_type_fn: Callable[[object], computation_types.Type],
  ):
    """Crates a polymorphic function with a given function factory.

    Args:
      concrete_function_factory: A callable that accepts a (non-None) type as an
        argument, as well as an optional boolean `unpack` argument which should
        be treated as documented in
        `function_utils.wrap_as_zero_or_one_arg_callable`. The callable must
        return a `Computation` instance that's been created to accept a single
        positional argument of this type (to be reused for future calls with
        parameters of a matching type).
      infer_type_fn: A `Callable` used to convert a backend-specific value to a
        `federated_language.Type`.
    """
    self._concrete_function_factory = concrete_function_factory
    self._infer_type_fn = infer_type_fn
    self._concrete_function_cache = {}

  def fn_for_argument_type(
      self, arg_type: computation_types.Type, unpack: Optional[bool] = None
  ) -> computation_impl.ConcreteComputation:
    """Concretizes this function with the provided `arg_type`.

    The first time this function is called with a particular type on a
    given `PolymorphicComputation` (or this `PolymorphicComputation` is called
    with an argument of the given type), the underlying function will be
    traced using the provided argument type as input. Later calls will
    return the cached computed concrete function.

    Args:
      arg_type: The argument type to use when concretizing this function.
      unpack: Whether to force unpacking the arguments (`True`), never unpack
        the arguments (`False`), or infer whether or not to unpack the arguments
        (`None`).

    Returns:
      The `federated_language.framework.ConcreteComputation` that results from
      tracing this
      `PolymorphicComputation` with `arg_type.
    """
    key = repr(arg_type) + str(unpack)
    concrete_fn = self._concrete_function_cache.get(key)
    if not concrete_fn:
      concrete_fn = (self._concrete_function_factory)(arg_type, unpack)
      if concrete_fn.type_signature.parameter != arg_type:
        raise TypeError(
            'Expected a concrete function that takes parameter {}, got one '
            'that takes {}.'.format(
                arg_type, concrete_fn.type_signature.parameter
            )
        )
      self._concrete_function_cache[key] = concrete_fn
    return concrete_fn

  def __call__(self, *args, **kwargs):
    """Invokes this polymorphic function with a given set of arguments.

    Args:
      *args: Positional args.
      **kwargs: Keyword args.

    Returns:
      The result of calling a concrete function, instantiated on demand based
      on the argument types (and cached for future calls).

    Raises:
      TypeError: if the concrete functions created by the factory are of the
        wrong computation_types.
    """
    packed_arg = function_utils.pack_args_into_struct(args, kwargs)
    args_type = self._infer_type_fn(args)
    if not isinstance(args_type, computation_types.StructType):
      raise ValueError
    kwargs_type = self._infer_type_fn(kwargs)
    if not isinstance(kwargs_type, computation_types.StructType):
      raise ValueError
    arg_type = computation_types.StructType([
        *args_type.items(),
        *kwargs_type.items(),
    ])
    # We know the argument types have been packed, so force unpacking.
    concrete_fn = self.fn_for_argument_type(arg_type, unpack=True)
    return concrete_fn(packed_arg)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A base Python interface for all types of executors."""

import abc
from typing import Optional

from federated_language.executor import executor_value_base


class RetryableError(Exception):
  """Raised when execution fails and can be retried."""


class Executor(abc.ABC):
  """Represents the abstract interface that all executors must implement."""

  @abc.abstractmethod
  def close(self):
    """Release resources associated with this Executor, if any.

    If the executor has one or more target Executors, implementation of this
    method must close them.
    """
    raise NotImplementedError

  @abc.abstractmethod
  async def create_value(
      self, value, type_spec=None
  ) -> executor_value_base.ExecutorValue:
    """A coroutine that creates embedded value from `value` of type `type_spec`.

    This function is used to embed a value within the executor. The argument
    can be one of the plain Python types, a nested structure, a representation
    of a computation, etc. Once embedded, the value can be further passed
    around within the executor. For functional values, embedding them prior to
    invocation potentially allows the executor to amortize overhead across
    multiple calls.

    Args:
      value: An object that represents the value to embed within the executor.
      type_spec: An optional `federated_language.Type` of the value represented
        by this object, or something convertible to it. The type can only be
        omitted if the value is a instance of `federated_language.TypedObject`.

    Returns:
      An instance of `ExecutorValue` that represents the embedded value.
    """
    raise NotImplementedError

  @abc.abstractmethod
  async def create_call(
      self,
      comp: executor_value_base.ExecutorValue,
      arg: Optional[executor_value_base.ExecutorValue] = None,
  ) -> executor_value_base.ExecutorValue:
    """A coroutine that creates a call to `comp` with optional argument `arg`.

    Args:
      comp: The computation to invoke. It must have been first embedded in the
        executor by calling `create_value()` on it first.
      arg: An optional argument of the call, or `None` if no argument was
        supplied. If it is present, it must have been embedded in the executor
        by calling `create_value()` on it first.

    Returns:
      An instance of `ExecutorValue` that represents the constructed call.
    """
    raise NotImplementedError

  @abc.abstractmethod
  async def create_struct(self, elements) -> executor_value_base.ExecutorValue:
    """A coroutine that creates a tuple of `elements`.

    Args:
      elements: A collection of `ExecutorValue`s to create a tuple from. The
        collection may be of any kind accepted by `structure.from_container`,
        including dictionaries and lists. The `ExecutorValues` in the container
        must have been created by calling `create_value` on this executor.

    Returns:
      An instance of `ExecutorValue` that represents the constructed tuple.
    """
    raise NotImplementedError

  @abc.abstractmethod
  async def create_selection(
      self, source, index
  ) -> executor_value_base.ExecutorValue:
    """A coroutine that creates a selection from `source`.

    Args:
      source: The source to select from. The source must have been embedded in
        this executor by invoking `create_value()` on it first.
      index: An integer index to select.

    Returns:
      An instance of `ExecutorValue` that represents the constructed selection.
    """
    raise NotImplementedError



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: federated_language/proto/data_type.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'federated_language/proto/data_type.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(federated_language/proto/data_type.proto\x12\x12\x66\x65\x64\x65rated_language*\xa6\x02\n\x08\x44\x61taType\x12\x0e\n\nDT_INVALID\x10\x00\x12\x0b\n\x07\x44T_BOOL\x10\n\x12\x0b\n\x07\x44T_INT8\x10\x06\x12\x0c\n\x08\x44T_INT16\x10\x05\x12\x0c\n\x08\x44T_INT32\x10\x03\x12\x0c\n\x08\x44T_INT64\x10\t\x12\x0c\n\x08\x44T_UINT8\x10\x04\x12\r\n\tDT_UINT16\x10\x11\x12\r\n\tDT_UINT32\x10\x16\x12\r\n\tDT_UINT64\x10\x17\x12\x0b\n\x07\x44T_HALF\x10\x13\x12\x0c\n\x08\x44T_FLOAT\x10\x01\x12\r\n\tDT_DOUBLE\x10\x02\x12\x10\n\x0c\x44T_COMPLEX64\x10\x08\x12\x11\n\rDT_COMPLEX128\x10\x12\x12\x0f\n\x0b\x44T_BFLOAT16\x10\x0e\x12\r\n\tDT_STRING\x10\x07\"\x04\x08\x0b\x10\x0b\"\x04\x08\x0c\x10\x0c\"\x04\x08\r\x10\r\"\x04\x08\x0f\x10\x0f\"\x04\x08\x10\x10\x10\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'federated_language.proto.data_type_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_DATATYPE']._serialized_start=65
  _globals['_DATATYPE']._serialized_end=359
# @@protoc_insertion_point(module_scope)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Container for structures with named and/or unnamed fields."""

import collections
from collections.abc import Callable, Iterable, Iterator, Mapping
import typing
from typing import Generic, Optional, TypeVar, Union

import attrs
from federated_language.common_libs import py_typecheck
import tree


_T = TypeVar('_T')
_U = TypeVar('_U')


class Struct(Generic[_T]):
  """Represents a struct-like structure with named and/or unnamed fields.

  `Struct`s are similar to `collections.namedtuple` in that their elements can
  be accessed by name or by index. However, `Struct`s provide a performance
  improvement over `collections.namedtuple` by using a single class to
  represent values with many different possible structures, rather than
  creating a brand new class for every new instance.

  `Struct`s are commonly used inside Tensorflow Federated as a standard
  intermediate representation of other structure types, including `list`s,
  `tuple`s, `dict`s, `namedtuple`s, and `attr.s` classes.

  Example:

  ```python
  x = Struct([('foo', 10), (None, 20), ('bar', 30)])

  len(x) == 3
  x[0] == 10
  x[1] == 20
  x[2] == 30
  list(iter(x)) == [10, 20, 30]
  dir(x) == ['bar', 'foo']
  x.foo == 10
  x['bar'] == 30
  ```

  Note that field names are optional, allowing `Struct` to be used like an
  ordinary positional tuple.
  """

  __slots__ = (
      '_hash',
      '_element_array',
      '_name_to_index',
      '_name_array',
      '_elements_cache',
  )

  @classmethod
  def named(cls, **kwargs: _T) -> 'Struct':
    """Constructs a new `Struct` with all named elements."""
    return cls(tuple(kwargs.items()))

  @classmethod
  def unnamed(cls, *args: _T) -> 'Struct':
    """Constructs a new `Struct` with all unnamed elements."""
    return cls(tuple((None, v) for v in args))

  def __init__(self, elements: Iterable[tuple[Optional[str], _T]]):
    """Constructs a new `Struct` with the given elements.

    Args:
      elements: An iterable of element specifications, each being a pair
        consisting of the element name (either `str`, or `None`), and the
        element value. The order is significant.

    Raises:
      TypeError: if the `elements` are not a list, or if any of the items on
        the list is not a pair with a string at the first position.
    """
    values = []
    names = []
    name_to_index = {}
    reserved_names = frozenset(('_asdict',) + Struct.__slots__)
    for idx, e in enumerate(elements):
      if not py_typecheck.is_name_value_pair(e):
        raise TypeError(
            'Expected every item on the list to be a pair in which the first '
            'element is a string, found {!r}.'.format(e)
        )
      name, value = e
      if name in reserved_names:
        raise ValueError(
            'The names in {} are reserved. You passed the name {}.'.format(
                reserved_names, name
            )
        )
      elif name in name_to_index:
        raise ValueError(
            '`Struct` does not support duplicated names, found {}.'.format(
                [e[0] for e in elements]
            )
        )
      names.append(name)
      values.append(value)
      if name is not None:
        name_to_index[name] = idx
    self._element_array = tuple(values)
    self._name_to_index = name_to_index
    self._name_array = names
    self._hash = None
    self._elements_cache = None

  def _elements(self) -> list[tuple[Optional[str], _T]]:
    if self._elements_cache is None:
      self._elements_cache = list(zip(self._name_array, self._element_array))
    return self._elements_cache

  def __len__(self) -> int:
    return len(self._element_array)

  def __iter__(self) -> Iterator[_T]:
    return iter(self._element_array)

  def __dir__(self) -> list[str]:
    """The list of names.

    IMPORTANT: `len(self)` may be greater than `len(dir(self))`, since field
    names are not required by `Struct`.

    IMPORTANT: the Python `dir()` built-in sorts the list returned by this
    method.

    Returns:
      A `list` of `str`.
    """
    return list(self._name_to_index.keys())

  def __getitem__(self, key: Union[int, str, slice]) -> _T:
    if isinstance(key, str):
      return self.__getattr__(key)
    if isinstance(key, int):
      if key < 0 or key >= len(self._element_array):
        raise IndexError(
            'Element index {} is out of range, `Struct` has {} elements.'
            .format(key, len(self._element_array))
        )
    return self._element_array[key]

  def __getattr__(self, name: str) -> _T:
    if name not in self._name_to_index:
      raise AttributeError(
          'The `Struct` of length {:d} does not have named field "{!s}". '
          'Fields (up to first 10): {!s}'.format(
              len(self._element_array),
              name,
              list(self._name_to_index.keys())[:10],
          )
      )
    return self._element_array[self._name_to_index[name]]

  def __setstate__(self, state):
    # IMPORTANT: At unpickling time, `__getattr__()` is called by the
    # implementation of `object.__reduce__()`. However, `__init__()` is not
    # called when unpickling an instance. This class implements `__getattr__()`
    # and accesses attributes that have not been set. Implementing
    # `__setstate__` instead of relying on `object.__reduce__()` prevents this.
    dict_attributes, slot_attributes = state
    self.__dict__.update(dict_attributes)
    for name, value in slot_attributes.items():
      setattr(self, name, value)

  def __eq__(self, other: object) -> bool:
    if self is other:
      return True
    elif not isinstance(other, Struct):
      return NotImplemented
    return (
        self._element_array,
        self._name_array,
    ) == (
        other._element_array,
        other._name_array,
    )

  def __ne__(self, other: object) -> bool:
    return not self == other

  def __repr__(self) -> str:
    return 'Struct([{}])'.format(
        ', '.join('({!r}, {!r})'.format(n, v) for n, v in iter_elements(self))
    )

  def __str__(self) -> str:
    def _element_str(element: tuple[Optional[str], object]) -> str:
      name, value = element
      if name is not None:
        return '{}={}'.format(name, value)
      return str(value)

    return '<{}>'.format(','.join(_element_str(e) for e in iter_elements(self)))

  def __hash__(self) -> int:
    if self._hash is None:
      self._hash = hash((
          'Struct',  # salting to avoid type mismatch.
          self._element_array,
          tuple(self._name_array),
      ))
    return self._hash

  def _asdict(
      self, recursive: bool = False
  ) -> collections.OrderedDict[str, _T]:
    """Returns an `collections.OrderedDict` mapping field names to their values.

    Args:
      recursive: Whether to convert nested `Struct`s recursively.
    """
    return to_odict(self, recursive=recursive)


def name_list(struct: Struct) -> list[str]:
  """Returns a `list` of the names of the named fields in `struct`.

  Args:
    struct: An instance of `Struct`.

  Returns:
    The list of string names for the fields that are named. Names appear in
    order, skipping names that are `None`.
  """
  names = struct._name_array  # pylint: disable=protected-access
  return [n for n in names if n is not None]


def name_list_with_nones(struct: Struct) -> list[Optional[str]]:
  """Returns an iterator over the names of all fields in `struct`."""
  return struct._name_array  # pylint: disable=protected-access


def to_elements(struct: Struct[_T]) -> list[tuple[Optional[str], _T]]:
  """Retrieves the list of (name, value) pairs from a `Struct`.

  Modeled as a module function rather than a method of `Struct` to avoid
  naming conflicts with the tuple attributes, and so as not to expose the user
  to this implementation-oriented functionality.

  Args:
    struct: An instance of `Struct`.

  Returns:
    The list of (name, value) pairs in which names can be None. Identical to
    the format that's accepted by the tuple constructor.

  Raises:
    TypeError: if the argument is not an `Struct`.
  """
  return struct._elements().copy()  # pylint: disable=protected-access


def iter_elements(struct: Struct[_T]) -> Iterator[tuple[Optional[str], _T]]:
  """Returns an iterator over (name, value) pairs from a `Struct`.

  Modeled as a module function rather than a method of `Struct` to avoid
  naming conflicts with the tuple attributes, and so as not to expose the user
  to this implementation-oriented functionality.

  Args:
    struct: An instance of `Struct`.

  Returns:
    An iterator of 2-tuples of name, value pairs, representing the elements of
      `struct`.

  Raises:
    TypeError: if the argument is not an `Struct`.
  """
  return iter(struct._elements())  # pylint: disable=protected-access


def to_odict(
    struct: Struct[_T], recursive: bool = False
) -> collections.OrderedDict[str, _T]:
  """Returns `struct` as an `collections.OrderedDict`, if possible.

  Args:
    struct: An `Struct`.
    recursive: Whether to convert nested `Struct`s recursively.

  Raises:
    ValueError: If the `Struct` contains unnamed elements.
  """

  def _to_odict(
      elements: list[tuple[Optional[str], _T]],
  ) -> collections.OrderedDict[str, _T]:
    for name, _ in elements:
      if name is None:
        raise ValueError(
            'Cannot convert an `Struct` with unnamed entries to a '
            '`collections.OrderedDict`: {}'.format(struct)
        )
    elements = typing.cast(list[tuple[str, _T]], elements)
    return collections.OrderedDict(elements)

  if recursive:
    return _to_container_recursive(struct, _to_odict)
  else:
    return _to_odict(to_elements(struct))


def to_odict_or_tuple(
    struct: Struct[_T], recursive: bool = True
) -> Union[collections.OrderedDict[str, _T], tuple[_T, ...]]:
  """Returns `struct` as an `collections.OrderedDict` or `tuple`, if possible.

  If all elements of `struct` have names, convert `struct` to an
  `collections.OrderedDict`. If no element has a name, convert `struct` to a
  `tuple`. If
  `struct` has both named and unnamed elements, raise an error.

  Args:
    struct: A `Struct`.
    recursive: Whether to convert nested `Struct`s recursively.

  Raises:
    ValueError: If `struct` (or any nested `Struct` when `recursive=True`)
      contains both named and unnamed elements.
  """

  def _to_odict_or_tuple(
      elements: list[tuple[Optional[str], _T]],
  ) -> Union[collections.OrderedDict[str, _T], tuple[_T, ...]]:
    fields_are_named = tuple(name is not None for name, _ in elements)
    if any(fields_are_named):
      if not all(fields_are_named):
        raise ValueError(
            'Cannot convert a `Struct` with both named and unnamed '
            'entries to an collections.OrderedDict or tuple: {!r}'.format(
                struct
            )
        )
      elements = typing.cast(list[tuple[str, _T]], elements)
      return collections.OrderedDict(elements)
    else:
      return tuple(value for _, value in elements)

  if recursive:
    return _to_container_recursive(struct, _to_odict_or_tuple)
  else:
    return _to_odict_or_tuple(to_elements(struct))


def flatten(struct: object) -> list[object]:
  """Returns a list of values in a possibly recursively nested `Struct`.

  NOTE: This implementation is not compatible with the approach of
  `tf.nest.flatten`, which enforces lexical order for
  `collections.OrderedDict`s.

  Args:
    struct: A `Struct`, possibly recursively nested, or a non-`Struct` element
      that can be packed with `tf.nest.flatten`. If `struct` has
      non-`Struct`-typed fields which should be flattened further, they should
      not contain inner `Structs`, as these will not be flattened (e.g.
      `Struct([('a', collections.OrderedDict(b=Struct([('c', 5)])))])` would not
      be valid).

  Returns:
    The list of leaf values in the `Struct`.
  """
  if not isinstance(struct, Struct):
    return tree.flatten(struct)
  else:
    result = []
    for _, v in iter_elements(struct):
      result.extend(flatten(v))
    return result


def pack_sequence_as(
    structure: Struct[_T], flat_sequence: list[_T]
) -> Struct[_T]:
  """Returns a list of values in a possibly recursively nested `Struct`.

  Args:
    structure: A `Struct`, possibly recursively nested.
    flat_sequence: A flat Python list of values.

  Returns:
    A `Struct` nested the same way as `structure`, but with leaves
    replaced with `flat_sequence` such that when flatten, it yields a list
    with the same contents as `flat_sequence`.
  """

  def _pack(
      structure, flat_sequence: list[_T], position: int
  ) -> tuple[Struct[_T], int]:
    """Pack a leaf element or recurvisely iterate over an `Struct`."""
    if not isinstance(structure, Struct):
      # Ensure that our leaf values are not structures.
      if isinstance(
          structure, (list, dict, py_typecheck.SupportsNamedTuple)
      ) or attrs.has(type(structure)):
        raise TypeError(
            'Cannot pack sequence into type {!s}, only structures of '
            '`Struct` are supported, found a structure with types '
            '{!s}).'.format(type(structure), structure)
        )

      return flat_sequence[position], position + 1
    else:
      elements = []
      for k, v in iter_elements(structure):
        packed_v, position = _pack(v, flat_sequence, position)
        elements.append((k, packed_v))
      return Struct(elements), position

  result, _ = _pack(structure, flat_sequence, 0)
  # NOTE: trailing elements are currently ignored.
  return result


def _is_same_structure(a: Struct, b: Struct) -> bool:
  """Compares whether `a` and `b` have the same nested structure.

  This method is analogous to `tf.nest.assert_same_structure`,
  but returns a boolean rather than throwing an exception.

  Args:
    a: a `Struct` object.
    b: a `Struct` object.

  Returns:
    True iff `a` and `b` have the same nested structure.

  Raises:
    TypeError: if `a` or `b` are not of type `Struct`.
  """
  elems_a = to_elements(a)
  elems_b = to_elements(b)
  if len(elems_a) != len(elems_b):
    return False
  for elem_a, elem_b in zip(elems_a, elems_b):
    val_a = elem_a[1]
    val_b = elem_b[1]
    if elem_a[0] != elem_b[0]:
      return False
    if isinstance(val_a, Struct) and isinstance(val_b, Struct):
      return _is_same_structure(val_a, val_b)
    elif isinstance(val_a, Struct) or isinstance(val_b, Struct):
      return False
    else:
      try:
        tree.assert_same_structure(val_a, val_b, check_types=True)
      except (ValueError, TypeError):
        return False
  return True


def map_structure(fn: Callable[..., object], *structures: Struct) -> object:
  """Applies `fn` to each entry in `structure` and returns a new structure.

  This is a special implementation of `tf.nest.map_structure`
  that works for `Struct`.

  Args:
    fn: a callable that accepts as many arguments as there are structures.
    *structures: a scalar, tuple, or list of constructed scalars and/or
      tuples/lists, or scalars. NOTE: numpy arrays are considered scalars.

  Returns:
    A new structure with the same arity as `structure` and same type as
    `structure[0]`, whose values correspond to `fn(x[0], x[1], ...)` where
    `x[i]` is a value in the corresponding location in `structure[i]`.

  Raises:
    TypeError: if `fn` is not a callable, or *structure is not all `Struct` or
      all `tf.Tensor` typed values.
    ValueError: if `*structure` is empty.
  """
  if not structures:
    raise ValueError('Must provide at least one structure')

  if not all(isinstance(s, Struct) for s in structures):
    return tree.map_structure(fn, *structures)

  for i, other in enumerate(structures[1:]):
    if not _is_same_structure(structures[0], other):
      raise TypeError(
          'Structure at position {} is not the same structure'.format(i)
      )

  flat_structure = [flatten(s) for s in structures]
  entries = zip(*flat_structure)
  s = [fn(*x) for x in entries]

  return pack_sequence_as(structures[0], s)


def from_container(value: object, recursive=False) -> Struct:
  """Creates an instance of `Struct` from a Python container.

  By default, this conversion is only performed at the top level for Python
  dictionaries, `collections.OrderedDict`s, `namedtuple`s, `list`s,
  `tuple`s, and `attr.s` classes. Elements of these structures are not
  recursively converted.

  Args:
    value: _The Python container to convert.
    recursive: Whether to convert elements recursively (`False` by default).

  Returns:
    The corresponding instance of `Struct`.

  Raises:
    TypeError: If the `value` is not of one of the supported container types.
  """

  def _convert(
      value: object, recursive: bool, must_be_container: bool = False
  ) -> Struct:
    """The actual conversion function.

    Args:
      value: Same as in `from_container`.
      recursive: Same as in `from_container`.
      must_be_container: When set to `True`, causes an exception to be raised if
        `value` is not a container.

    Returns:
      The result of conversion.

    Raises:
      TypeError: If `value` is not a container and `must_be_container` has been
        set to `True`.
    """
    # TODO: b/224484886 - Downcasting to all handled types.
    value = typing.cast(
        Union[
            Struct,
            py_typecheck.SupportsNamedTuple,
            Mapping[str, object],
            dict[str, object],
            tuple[object, ...],
            list[object],
        ],
        value,
    )
    if isinstance(value, Struct):
      if recursive:
        return Struct((k, _convert(v, True)) for k, v in iter_elements(value))
      else:
        return value
    elif attrs.has(type(value)):
      return _convert(
          attrs.asdict(value, recurse=False), recursive, must_be_container
      )
    elif isinstance(value, py_typecheck.SupportsNamedTuple):
      return _convert(value._asdict(), recursive, must_be_container)
    elif isinstance(value, Mapping):
      items = value.items()
      if recursive:
        return Struct((k, _convert(v, True)) for k, v in items)
      else:
        return Struct(items)
    elif isinstance(value, (tuple, list)):
      if recursive:
        return Struct((None, _convert(v, True)) for v in value)
      else:
        return Struct((None, v) for v in value)
    elif must_be_container:
      raise TypeError(
          f'Unable to convert a Python object of type {type(value)} into an'
          f' `Struct`. Object: {value}'
      )
    else:
      return value

  return _convert(value, recursive, must_be_container=True)


def _to_container_recursive(
    value: Struct[_T],
    container_fn: Callable[[list[tuple[Optional[str], _T]]], _U],
) -> _U:
  """Recursively converts the `Struct` `value` to a new container type.

  This function is always recursive, since the non-recursive version would be
  just `container_fn(value)`.

  NOTE: _This function will only recurse through `Struct`s, so if called
  on the input `Struct([('a', 1), ('b', {'c': Struct(...)})])`
  the inner `Struct` will not be converted, because we do not recurse
  through Python `dict`s.

  Args:
    value: An `Struct`, possibly nested.
    container_fn: A function that takes a `list` of `(name, value)` tuples ( the
      elements of an `Struct`), and returns a new container holding the same
      values.

  Returns:
    A nested container of the type returned by `container_fn`.
  """

  def recurse(v):
    if isinstance(v, Struct):
      return _to_container_recursive(v, container_fn)
    else:
      return v

  return container_fn([(k, recurse(v)) for k, v in iter_elements(value)])


def has_field(structure: Struct, field: str) -> bool:
  """Returns `True` if the `structure` has the `field`.

  Args:
    structure: An instance of `Struct`.
    field: A string, the field to test for.
  """
  return field in structure._name_array  # pylint: disable=protected-access


def update_struct(structure, **kwargs):
  """Constructs a new `structure` with new values for fields in `kwargs`.

  This is a helper method for working structured objects in a functional manner.
  This method will create a new structure where the fields named by keys in
  `kwargs` replaced with the associated values.

  NOTE: _This method only works on the first level of `structure`, and does not
  recurse in the case of nested structures. A field that is itself a structure
  can be replaced with another structure.

  Args:
    structure: _The structure with named fields to update.
    **kwargs: _The list of key-value pairs of fields to update in `structure`.

  Returns:
    A new instance of the same type of `structure`, with the fields named
    in the keys of `**kwargs` replaced with the associated values.

  Raises:
    KeyError: If kwargs contains a field that is not in structure.
    TypeError: If structure is not a structure with named fields.
  """
  if not isinstance(
      structure, (Struct, Mapping, py_typecheck.SupportsNamedTuple)
  ) and not attrs.has(type(structure)):
    raise TypeError(
        '`structure` must be a structure with named fields (e.g. '
        'dict, attrs class, collections.namedtuple, '
        'federated_language.structure.Struct), but '
        'found {}'.format(type(structure))
    )
  if isinstance(structure, Struct):
    elements = [
        (k, v) if k not in kwargs else (k, kwargs.pop(k))
        for k, v in iter_elements(structure)
    ]
    if kwargs:
      raise KeyError(f'`structure` does not contain fields named {kwargs}')
    return Struct(elements)
  elif isinstance(structure, py_typecheck.SupportsNamedTuple):
    # In Python 3.8 and later `_asdict` no longer return OrdereDict, rather a
    # regular `dict`, so we wrap here to get consistent types across Python
    # version.s
    dictionary = structure._asdict()
  elif attrs.has(type(structure)):
    dictionary = attrs.asdict(structure, recurse=False)
  else:
    for key in kwargs:
      if key not in structure:
        raise KeyError(
            'structure does not contain a field named "{!s}"'.format(key)
        )
    # Create a copy to prevent mutation of the original `structure`
    dictionary = type(structure)(**structure)
  dictionary.update(kwargs)
  if isinstance(structure, Mapping):
    return dictionary
  return type(structure)(**dictionary)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A context for execution based on an embedded executor instance."""

import asyncio
from collections.abc import Callable, Iterator, Mapping, Sequence
import contextlib
import functools
from typing import Optional

from federated_language.common_libs import retrying
from federated_language.common_libs import structure
from federated_language.common_libs import tracing
from federated_language.computation import computation_base
from federated_language.computation import function_utils
from federated_language.context_stack import context
from federated_language.executor import cardinalities_utils
from federated_language.executor import executor_base
from federated_language.executor import executor_factory  # pylint: disable=unused-import
from federated_language.executor import executor_value_base
from federated_language.types import computation_types
from federated_language.types import placements
from federated_language.types import type_conversions
import tree


def _is_retryable_error(exception: Exception) -> bool:
  return isinstance(exception, executor_base.RetryableError)


async def _ingest(
    executor: executor_base.Executor,
    val,
    type_spec: computation_types.Type,
) -> executor_value_base.ExecutorValue:
  """A coroutine that handles ingestion.

  Args:
    executor: An instance of `executor_base.Executor`.
    val: The first argument to `AsyncExecutionContext.ingest()`.
    type_spec: The second argument to `AsyncExecutionContext.ingest()`.

  Returns:
    The result of the ingestion.

  Raises:
    TypeError: If the `val` is not a value of type `type_spec`.
  """
  if isinstance(val, executor_value_base.ExecutorValue):
    return val
  elif isinstance(val, structure.Struct) and not isinstance(
      type_spec, computation_types.FederatedType
  ):
    if not isinstance(type_spec, computation_types.StructType):
      raise ValueError(
          f'Expected a `federated_language.StructType`, found {type_spec}.'
      )
    v_elem = structure.to_elements(val)
    t_elem = list(type_spec.items())
    if len(v_elem) != len(t_elem):
      raise TypeError(
          'Value {} does not match type {}: mismatching tuple length.'.format(
              val, type_spec
          )
      )
    for (vk, _), (tk, _) in zip(v_elem, t_elem):
      if vk not in [tk, None]:
        raise TypeError(
            'Value {} does not match type {}: mismatching tuple element '
            'names {} vs. {}.'.format(val, type_spec, vk, tk)
        )
    ingested = []
    for (_, v), (_, t) in zip(v_elem, t_elem):
      ingested.append(_ingest(executor, v, t))
    ingested = await asyncio.gather(*ingested)
    return await executor.create_struct(
        structure.Struct(
            (name, val) for (name, _), val in zip(t_elem, ingested)
        )
    )
  else:
    return await executor.create_value(val, type_spec)


async def _invoke(
    executor: executor_base.Executor,
    comp,
    arg: Optional[executor_value_base.ExecutorValue],
    result_type: computation_types.Type,
) -> object:
  """A coroutine that handles invocation.

  Args:
    executor: An instance of `executor_base.Executor`.
    comp: The first argument to `AsyncExecutionContext.invoke()`.
    arg: The optional second argument to `AsyncExecutionContext.invoke()`.
    result_type: The type signature of the result. This is used to convert the
      execution result into the proper container types.

  Returns:
    The result of the invocation.
  """
  comp = await executor.create_value(
      comp,
      comp.type_signature,  # pytype: disable=attribute-error
  )
  result = await executor.create_call(comp, arg)
  result_value = await result.compute()
  return type_conversions.type_to_py_container(result_value, result_type)


class AsyncExecutionContext(context.AsyncContext):
  """An asynchronous execution context backed by an `executor_base.Executor`.

  This context's `ingest` and `invoke` methods return Python coroutine objects
  which represent the actual work of ingestion and invocation in the backing
  executor.

  This context will support concurrent invocation of multiple computations if
  their arguments have the same cardinalities.
  """

  def __init__(
      self,
      executor_fn: executor_factory.ExecutorFactory,
      compiler_fn: Optional[
          Callable[[computation_base.Computation], object]
      ] = None,
      *,
      transform_args: Optional[Callable[[object], object]] = None,
      transform_result: Optional[Callable[[object], object]] = None,
      cardinality_inference_fn: cardinalities_utils.CardinalityInferenceFnType = cardinalities_utils.infer_cardinalities,
  ):
    """Initializes an execution context.

    Args:
      executor_fn: Instance of `executor_factory.ExecutorFactory`.
      compiler_fn: A Python function that will be used to compile a computation.
      transform_args: An `Optional` `Callable` used to transform the args before
        they are passed to the computation.
      transform_result: An `Optional` `Callable` used to transform the result
        before it is returned.
      cardinality_inference_fn: A Python function specifying how to infer
        cardinalities from arguments (and their associated types). The value
        returned by this function will be passed to the `create_executor` method
        of `executor_fn` to construct a `federated_language.framework.Executor`
        instance.
    """
    super().__init__()
    self._executor_factory = executor_fn
    if compiler_fn is not None:
      cache_decorator = functools.lru_cache()
      self._compiler = cache_decorator(compiler_fn)
    else:
      self._compiler = None
    self._transform_args = transform_args
    self._transform_result = transform_result
    self._cardinality_inference_fn = cardinality_inference_fn

  @property
  def executor_factory(self) -> executor_factory.ExecutorFactory:
    return self._executor_factory

  @contextlib.contextmanager
  def _reset_factory_on_error(
      self,
      cardinalities: Mapping[placements.PlacementLiteral, int],
  ) -> Iterator[executor_base.Executor]:
    try:
      # We pass a copy down to prevent the caller from mutating.
      yield self._executor_factory.create_executor({**cardinalities})
    except Exception:
      self._executor_factory.clean_up_executor({**cardinalities})
      raise

  @retrying.retry(
      retry_on_exception_filter=_is_retryable_error,
      wait_max_ms=30 * 1000,
      wait_multiplier=2,
  )
  async def invoke(self, comp, arg):
    if asyncio.iscoroutine(arg):
      # Awaiting if we are passed a coro allows us to install and use the async
      # context in conjunction with ConcreteComputations' implementation of
      # __call__.
      arg = await arg

    if not isinstance(comp.type_signature, computation_types.FunctionType):
      raise ValueError(
          'Expected a `federated_language.FunctionType`, found'
          f' {comp.type_signature}.'
      )

    if arg is not None and self._transform_args is not None:
      # `transform_args` is not intended to handle `structure.Struct`. Normalize
      # to a Python structure to make it simpler to handle; `args` is sometimes
      # a `structure.Struct` and sometimes it is not, other times it is a Python
      # structure that contains a `structure.Struct`.
      def _to_python(obj):
        if isinstance(obj, structure.Struct):
          return structure.to_odict_or_tuple(obj)
        else:
          return None

      if isinstance(arg, structure.Struct):
        args, kwargs = function_utils.unpack_args_from_struct(arg)
        args = tree.traverse(_to_python, args)
        args = self._transform_args(args)
        if not isinstance(args, Sequence):
          raise ValueError(
              f'Expected `args` to be a `Sequence`, found {type(args)}'
          )
        kwargs = tree.traverse(_to_python, kwargs)
        kwargs = self._transform_args(kwargs)
        if not isinstance(kwargs, Mapping):
          raise ValueError(
              f'Expected `kwargs` to be a `Mapping`, found {type(kwargs)}'
          )
        arg = function_utils.pack_args_into_struct(args, kwargs)
      else:
        arg = tree.traverse(_to_python, arg)
        arg = self._transform_args(arg)

    # Save the type signature before compiling. Compilation currently loses
    # container types, so we must remember them here so that they can be
    # restored in the output.
    result_type = comp.type_signature.result
    if self._compiler is not None:
      with tracing.span('ExecutionContext', 'Compile', span=True):
        comp = self._compiler(comp)

    with tracing.span('ExecutionContext', 'Invoke', span=True):
      if arg is not None:
        cardinalities = self._cardinality_inference_fn(
            arg,
            comp.type_signature.parameter,  # pytype: disable=attribute-error
        )
      else:
        cardinalities = {}

      with self._reset_factory_on_error(cardinalities) as executor:

        if arg is not None:
          arg = await tracing.wrap_coroutine_in_current_trace_context(
              _ingest(
                  executor,
                  arg,
                  comp.type_signature.parameter,  # pytype: disable=attribute-error
              )
          )

        result = await tracing.wrap_coroutine_in_current_trace_context(
            _invoke(executor, comp, arg, result_type)
        )

        if self._transform_result is not None:
          result = self._transform_result(result)
        return result



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for cardinality inference and handling."""

from collections.abc import Callable, Mapping, Sequence

from federated_language.types import computation_types
from federated_language.types import placements


def _merge_cardinalities(
    existing: Mapping[placements.PlacementLiteral, int],
    update: Mapping[placements.PlacementLiteral, int],
) -> Mapping[placements.PlacementLiteral, int]:
  """Returns the merged cardinalities after checking for conflicts."""
  if not update:
    return existing
  elif not existing:
    return update

  for k, v in update.items():
    if k in existing and existing[k] != v:
      raise ValueError(
          f'Conflicting cardinalities for {k}: {v} vs {existing[k]}'
      )

  cardinalities = {}
  cardinalities.update(existing)
  cardinalities.update(update)
  return cardinalities


class InvalidNonAllEqualValueError(TypeError):

  def __init__(self, value, type_spec):
    message = (
        f'Expected non-all-equal value with placement {type_spec.placement} '
        'to be a `list` or `tuple`, found a value of Python type '
        f'{type(value)}:\n{value}'
    )
    super().__init__(message)


# We define this type here to avoid having to redeclare it wherever we
# parameterize by a cardinality inference fn.
CardinalityInferenceFnType = Callable[
    [object, computation_types.Type],
    Mapping[placements.PlacementLiteral, int],
]


def infer_cardinalities(
    value: object, type_spec: computation_types.Type
) -> dict[placements.PlacementLiteral, int]:
  """Infers cardinalities from Python `value`.

  Allows for any Python object to represent a federated value; enforcing
  particular representations is not the job of this inference function, but
  rather ingestion functions lower in the stack.

  Args:
    value: Python object from which to infer placement cardinalities.
    type_spec: The type spec for `value`, determining the semantics for
      inferring cardinalities. That is, we only pull the cardinality off of
      federated types.

  Returns:
    Dict of cardinalities.

  Raises:
    ValueError: If conflicting cardinalities are inferred from `value`.
    TypeError: If the arguments are of the wrong types, or if `type_spec` is
      a federated type which is not `all_equal` but the yet-to-be-embedded
      `value` is not represented as a Python `list`.
  """
  if value is None:
    return {}

  if isinstance(type_spec, computation_types.FederatedType):
    if type_spec.all_equal:
      return {}
    if not isinstance(value, Sequence):
      raise InvalidNonAllEqualValueError(value, type_spec)
    return {type_spec.placement: len(value)}
  elif isinstance(type_spec, computation_types.StructType):
    cardinalities = {}

    if isinstance(value, Mapping):
      elements = value.values()
    else:
      elements = value

    for element, (_, element_type) in zip(elements, type_spec.items()):
      update = infer_cardinalities(element, element_type)
      cardinalities = _merge_cardinalities(cardinalities, update)
    return cardinalities
  else:
    return {}



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines context interfaces which evaluates computation invocations.

Invocations of Federated Language computations need to be treated differently
depending on the context in which they are invoked. For example:

*   During top-level Python simulations, computation invocations result in the
    computation being serialized and evaluated by the TensorFlow native runtime.

*   In functions decorated with `@federated_language.tensorflow.computation`,
computation
    invocations must import the body of the invoked function into the current
    TensorFlow graph.

Code can customize the way in which each of these calls are evaluated by setting
a specific context using a global or thread-local context stack.
"""

import abc
from typing import Any


class ContextError(RuntimeError):  # pylint: disable=g-bad-exception-name
  pass


class SyncContext(abc.ABC):
  """A synchronous context to evaluate of computations."""

  @abc.abstractmethod
  def invoke(self, comp: Any, arg: Any) -> Any:
    """Invokes computation `comp` with argument `arg`.

    Args:
      comp: The computation being invoked. The Python type of `comp` expected
        here (e.g., `pb.Computation`.
        `federated_language.framework.ConcreteComputation`, or other) may depend
        on the context. It is the responsibility of the concrete implementation
        of this interface to verify that the type of `comp` matches what the
        context is expecting.
      arg: The argument passed to the computation. If no argument is passed,
        this will be `None`. Structural argument types will be normalized into
        `federated_language.structure.Struct`s.

    Returns:
      The result of invocation, which is context-dependent.
    """
    raise NotImplementedError


class AsyncContext(abc.ABC):
  """An asynchronous context to evaluate of computations."""

  @abc.abstractmethod
  async def invoke(self, comp: Any, arg: Any) -> Any:
    """Invokes computation `comp` with argument `arg`.

    Args:
      comp: The computation being invoked. The Python type of `comp` expected
        here (e.g., `pb.Computation`.
        `federated_language.framework.ConcreteComputation`, or other) may depend
        on the context. It is the responsibility of the concrete implementation
        of this interface to verify that the type of `comp` matches what the
        context is expecting.
      arg: The argument passed to the computation. If no argument is passed,
        this will be `None`. Structural argument types will be normalized into
        `federated_language.structure.Struct`s.

    Returns:
      The result of invocation, which is context-dependent.
    """
    raise NotImplementedError



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A base Python interface for values embedded in executors."""

import abc

from federated_language.types import typed_object


class ExecutorValue(typed_object.TypedObject, abc.ABC):
  """Represents the abstract interface for values embedded within executors.

  The embedded values may represent computations in-flight that may materialize
  in the future or fail before they materialize.
  """

  @property
  @abc.abstractmethod
  def reference(self):
    """Returns a reference to the value without transferring ownership.

    A reference is an opaque object that is understood by the executors that
    produced the value, therefore:

    1. Executors need to preserve this contract in their implementation.
    2. Users of Executors should not need to depend on this value.
    """
    raise NotImplementedError

  @abc.abstractmethod
  async def compute(self):
    """A coroutine that asynchronously returns the computed form of the value.

    The computed form of a value can take a number of forms, such as primitive
    types in Python, numpy arrays, or even eager tensors in case this is an
    eager executor, or an executor backed by an eager one.

    Returns:
      The computed form of the value, as defined above.
    """
    raise NotImplementedError



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines an abstract interface for representing a federated context."""

import abc
from typing import Optional, Union

from federated_language.computation import computation_base
from federated_language.context_stack import context
from federated_language.context_stack import context_stack_impl
from federated_language.program import structure_utils
from federated_language.program import value_reference
from federated_language.types import computation_types
from federated_language.types import placements
from federated_language.types import type_analysis


ComputationArg = Union[
    value_reference.MaterializableStructure,
    object,
    computation_base.Computation,
]


def contains_only_server_placed_data(
    type_signature: computation_types.Type,
) -> bool:
  """Determines if `type_signature` contains only server-placed data.

  Determines if `type_signature` contains only:
  * `federated_language.StructType`s
  * `federated_language.SequenceType`s
  * server-placed `federated_language.FederatedType`s
  * `federated_language.TensorType`s

  Args:
      type_signature: The `federated_language.Type` to test.

  Returns:
    `True` if `type_signature` contains only server-placed data, otherwise
    `False`.
  """

  def _predicate(type_spec: computation_types.Type) -> bool:
    return isinstance(
        type_spec,
        (
            computation_types.StructType,
            computation_types.SequenceType,
            computation_types.TensorType,
        ),
    ) or (
        isinstance(type_spec, computation_types.FederatedType)
        and type_spec.placement is placements.SERVER
    )

  return type_analysis.contains_only(type_signature, _predicate)


class FederatedContext(context.SyncContext):
  """An abstract interface representing a federated context.

  A federated context supports invoking a limited set of
  `federated_language.Computation`s,
  making guarantees about what a `federated_language.Computation` can accept as
  an argument and
  what it returns when invoked.

  ## Restrictions on the Federated Language Type

  Arguments can be nested structures of values corresponding to the TensorFlow
  Federated type signature of the `federated_language.Computation`:

  * Server-placed values must be represented by
    `federated_language.program.MaterializableStructure`.
  * Client-placed values must be represented by structures of values returned by
    a `federated_language.program.FederatedDataSourceIterator`.

  Return values can be structures of
  `federated_language.program.MaterializableValueReference`s
  or a single `federated_language.program.MaterializableValueReference`, where a
  reference
  corresponds to the tensor-type of the Federated Language type signature in
  the return value of the invoked `federated_language.Computation`.

  ## Federated Language Type to Python Representation

  In order to interact with the value returned by a
  `federated_language.Computation`, it is
  helpful to be able to reason about the Python type of this value. In some way
  this Python type must depend on the Federated Language type signature of the
  associated value. To provide uniformity of experience and ease of reasoning,
  we specify the Python representation of values in a manner that can be stated
  entirely in the Federated Language typesystem.

  We have chosen to limit the Federated Language type signatures of invoked
  `federated_language.Computation`s to disallow the returning of client-placed
  values,
  `federated_language.SequenceTypes`, and `federated_language.FunctionTypes`, in
  order to reduced the area
  which needs to be supported by federated programs. Below we describe the
  mapping between Federated Language type signatures and Python
  representations of values that can be passed as arguments to or returned as
  results from `federated_language.Computation`s.

  Python representations of values that can be *accepted as an arguments to* or
  *returned as a value from* a `federated_language.Computation`:

  | Federated Language Type  | Python Representation                      |
  | ------------------------ | ------------------------------------------ |
  | `TensorType`             | `MaterializableValueReference`             |
  | `SequenceType`           | `MaterializableValueReference`             |
  | `FederatedType`          | Python representation of the `member` of   |
  : (server-placed)          : the `federated_language.FederatedType`     :
  | `StructWithPythonType`   | Python container of the                    |
  :                          : `federated_language.StructWithPythonType`  :
  | `StructType` (with no    | `collections.OrderedDict`                  |
  : Python type, all fields  :                                            :
  : named)                   :                                            :
  | `StructType` (with no    | `tuple`                                    |
  : Python type, no fields   :                                            :
  : named)                   :                                            :

  Python representations of values that can be only be *accepted as an arguments
  to* a `federated_language.Computation`:

  | Federated Language Type  | Python Representation                   |
  | ------------------------ | --------------------------------------- |
  | `FederatedType`          | Opaque object returned by               |
  : (client-placed)          : `DataSourceIterator.select`             :
  | `FunctionType`           | `federated_language.Computation`        |
  """

  @abc.abstractmethod
  def invoke(
      self,
      comp: computation_base.Computation,
      arg: Optional[ComputationArg],
  ) -> structure_utils.Structure[value_reference.MaterializableValueReference]:
    """Invokes the `comp` with the argument `arg`.

    Args:
      comp: The `federated_language.Computation` being invoked.
      arg: The optional argument of `comp`; server-placed values must be
        represented by `federated_language.program.MaterializableStructure`, and
        client-placed values must be represented by structures of values
        returned by a `federated_language.program.FederatedDataSourceIterator`.

    Returns:
      The result of invocation; a structure of
      `federated_language.program.MaterializableValueReference`.

    Raises:
      ValueError: If the result type of `comp` does not contain only structures,
      server-placed values, or tensors.
    """
    raise NotImplementedError


def check_in_federated_context() -> None:
  """Checks if the current context is a `federated_language.program.FederatedContext`."""
  context_stack = context_stack_impl.get_context_stack()
  if not isinstance(context_stack.current, FederatedContext):
    raise ValueError(
        'Expected the current context to be a'
        ' `federated_language.program.FederatedContext`, found'
        f' {type(context_stack.current)}.'
    )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for working with arrays."""

from typing import Optional, Union

from federated_language.proto import array_pb2
from federated_language.types import array_shape
from federated_language.types import dtype_utils
import ml_dtypes
import numpy as np

# Array is the Python representation of the `Array` protobuf, and is the native
# representation of an array.
Array = Union[
    # Python types
    bool,
    int,
    float,
    complex,
    str,
    bytes,
    # Numpy types
    np.generic,
    np.ndarray,
]


def from_proto(array_pb: array_pb2.Array) -> Array:
  """Returns an `Array` for the `array_pb`."""
  dtype = dtype_utils.from_proto(array_pb.dtype)
  shape = array_shape.from_proto(array_pb.shape)

  if dtype is np.bool_:
    value = array_pb.bool_list.value
  elif dtype is np.int8:
    value = array_pb.int8_list.value
  elif dtype is np.int16:
    value = array_pb.int16_list.value
  elif dtype is np.int32:
    value = array_pb.int32_list.value
  elif dtype is np.int64:
    value = array_pb.int64_list.value
  elif dtype is np.uint8:
    value = array_pb.uint8_list.value
  elif dtype is np.uint16:
    value = array_pb.uint16_list.value
  elif dtype is np.uint32:
    value = array_pb.uint32_list.value
  elif dtype is np.uint64:
    value = array_pb.uint64_list.value
  elif dtype is np.float16:
    value = array_pb.float16_list.value
    # Values of dtype `np.float16` are packed to and unpacked from a protobuf
    # field of type `int32` using the following logic in order to maintain
    # compatibility with how other external environments (e.g., TensorFlow, JAX)
    # represent values of `np.float16`.
    value = np.asarray(value, np.uint16).view(np.float16).tolist()
  elif dtype is np.float32:
    value = array_pb.float32_list.value
  elif dtype is np.float64:
    value = array_pb.float64_list.value
  elif dtype is np.complex64:
    if len(array_pb.complex64_list.value) % 2 != 0:
      raise ValueError(
          'Expected the number of complex values to be even, one real and one'
          ' imaginary part for each complex value.'
      )
    value = iter(array_pb.complex64_list.value)
    value = [complex(real, imag) for real, imag in zip(value, value)]
  elif dtype is np.complex128:
    if len(array_pb.complex128_list.value) % 2 != 0:
      raise ValueError(
          'Expected the number of complex values to be even, one real and one'
          ' imaginary part for each complex value.'
      )
    value = iter(array_pb.complex128_list.value)
    value = [complex(real, imag) for real, imag in zip(value, value)]
  elif dtype is ml_dtypes.bfloat16:
    value = array_pb.bfloat16_list.value
    # Values of dtype `ml_dtypes.bfloat16` are packed to and unpacked from a
    # protobuf field of type `int32` using the following logic in order to
    # maintain compatibility with how other external environments (e.g.,
    # TensorFlow, JAX) represent values of `ml_dtypes.bfloat16`.
    value = np.asarray(value, np.uint16).view(ml_dtypes.bfloat16).tolist()
  elif dtype is np.str_:
    value = array_pb.string_list.value
  else:
    raise NotImplementedError(f'Unexpected dtype found: {dtype}.')

  # Strings are stored as bytes in `array_pb2.Array` and trailing null values
  # are dropped when using `np.bytes_`, use `np.object_` instead.
  if dtype is np.str_:
    dtype = np.object_

  # `Array` is a `Union` of native Python types and numpy types. However, the
  # protobuf representation of `Array` contains additional information like
  # dtype and shape. This information is lost when returning native Python types
  # making it impossible to infer the original dtype later. Therefore, a numpy
  # value should almost always be returned from this function. String values are
  # an exception to this because it's not possible to represent null-terminated
  # scalar strings using numpy and this is ok because string types can only be
  # inferred as string types.
  if not array_shape.is_shape_scalar(shape):
    value = np.array(value, dtype).reshape(shape)
  else:
    (value,) = value
    value = dtype(value)

  return value


def to_proto(
    value: Array, *, dtype_hint: Optional[type[np.generic]] = None
) -> array_pb2.Array:
  """Returns an `array_pb2.Array` for the `value`."""

  if dtype_hint is not None:
    if not dtype_utils.is_valid_dtype(dtype_hint):
      raise ValueError(
          f'Expected `dtype_hint` to be a valid dtype, found {dtype_hint}.'
      )
    if not is_compatible_dtype(value, dtype_hint):
      raise ValueError(f'Expected {value} to be compatible with {dtype_hint}.')
    dtype = dtype_hint
  else:
    if isinstance(value, (np.generic, np.ndarray)):
      dtype = value.dtype.type
      # If the value has a dtype of `np.bytes_` or `np.object_`, the serialized
      # dtype should still be a `np.str_`.
      if np.issubdtype(dtype, np.bytes_) or np.issubdtype(dtype, np.object_):
        dtype = np.str_
    else:
      dtype = dtype_utils.infer_dtype(value)

  # Normalize to a numpy value; strings are stored as bytes in `array_pb2.Array`
  # and trailing null values are dropped when using `np.bytes_`, so use
  # `np.object_` instead.
  if dtype is np.str_:

    def _contains_type(value, classinfo):
      if isinstance(value, (np.generic, np.ndarray)):
        if value.size == 0:
          return False
        item = value.item(0)
      else:
        item = value
      return isinstance(item, classinfo)

    if _contains_type(value, str):
      value = np.asarray(value, np.bytes_)
    else:
      value = np.asarray(value, np.object_)
  else:
    value = np.asarray(value, dtype)

  dtype_pb = dtype_utils.to_proto(dtype)
  shape_pb = array_shape.to_proto(value.shape)
  value = value.flatten().tolist()

  if dtype is np.bool_:
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        bool_list=array_pb2.Array.BoolList(value=value),
    )
  elif dtype is np.int8:
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        int8_list=array_pb2.Array.IntList(value=value),
    )
  elif dtype is np.int16:
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        int16_list=array_pb2.Array.IntList(value=value),
    )
  elif dtype is np.int32:
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        int32_list=array_pb2.Array.IntList(value=value),
    )
  elif dtype is np.int64:
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        int64_list=array_pb2.Array.Int64List(value=value),
    )
  elif dtype is np.uint8:
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        uint8_list=array_pb2.Array.IntList(value=value),
    )
  elif dtype is np.uint16:
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        uint16_list=array_pb2.Array.IntList(value=value),
    )
  elif dtype is np.uint32:
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        uint32_list=array_pb2.Array.Uint32List(value=value),
    )
  elif dtype is np.uint64:
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        uint64_list=array_pb2.Array.Uint64List(value=value),
    )
  elif dtype is np.float16:
    # Values of dtype `np.float16` are packed to and unpacked from a protobuf
    # field of type `int32` using the following logic in order to maintain
    # compatibility with how other external environments (e.g., TensorFlow, JAX)
    # represent values of `np.float16`.
    value = np.asarray(value, np.float16).view(np.uint16).tolist()
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        float16_list=array_pb2.Array.IntList(value=value),
    )
  elif dtype is np.float32:
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        float32_list=array_pb2.Array.FloatList(value=value),
    )
  elif dtype is np.float64:
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        float64_list=array_pb2.Array.DoubleList(value=value),
    )
  elif dtype is np.complex64:
    packed_value = []
    for x in value:
      if not isinstance(x, complex):
        raise ValueError(f'Expected a complex type, found {type(x)}.')
      packed_value.extend([x.real, x.imag])
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        complex64_list=array_pb2.Array.FloatList(value=packed_value),
    )
  elif dtype is np.complex128:
    packed_value = []
    for x in value:
      if not isinstance(x, complex):
        raise ValueError(f'Expected a complex type, found {type(x)}.')
      packed_value.extend([x.real, x.imag])
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        complex128_list=array_pb2.Array.DoubleList(value=packed_value),
    )
  elif dtype is ml_dtypes.bfloat16:
    # Values of dtype `ml_dtypes.bfloat16` are packed to and unpacked from a
    # protobuf field of type `int32` using the following logic in order to
    # maintain compatibility with how other external environments (e.g.,
    # TensorFlow, JAX) represent values of `ml_dtypes.bfloat16`.
    value = np.asarray(value, ml_dtypes.bfloat16).view(np.uint16).tolist()
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        bfloat16_list=array_pb2.Array.IntList(value=value),
    )
  elif dtype is np.str_:
    return array_pb2.Array(
        dtype=dtype_pb,
        shape=shape_pb,
        string_list=array_pb2.Array.BytesList(value=value),
    )
  else:
    raise NotImplementedError(f'Unexpected dtype found: {dtype}.')


def from_proto_content(array_pb: array_pb2.Array) -> Array:
  """Returns an `Array` for the `array_pb`."""
  dtype = dtype_utils.from_proto(array_pb.dtype)
  shape = array_shape.from_proto(array_pb.shape)

  if dtype is not np.str_:
    value = np.frombuffer(array_pb.content, dtype)
  else:
    raise NotImplementedError(f'Unexpected dtype found: {dtype}.')

  # `Array` is a `Union` of native Python types and numpy types. However, the
  # protobuf representation of `Array` contains additional information like
  # dtype and shape. This information is lost when returning native Python types
  # making it impossible to infer the original dtype later. Therefore, a numpy
  # value should almost always be returned from this function. String values are
  # an exception to this because it's not possible to represent null-terminated
  # scalar strings using numpy and this is ok because string types can only be
  # inferred as string types.
  if not array_shape.is_shape_scalar(shape):
    value = value.reshape(shape)
  else:
    value = value.item()
    value = dtype(value)

  return value


def to_proto_content(
    value: Array, *, dtype_hint: Optional[type[np.generic]] = None
) -> array_pb2.Array:
  """Returns an `Array` for the `value`."""

  if dtype_hint is not None:
    if not dtype_utils.is_valid_dtype(dtype_hint):
      raise ValueError(
          f'Expected `dtype_hint` to be a valid dtype, found {dtype_hint}.'
      )
    if not is_compatible_dtype(value, dtype_hint):
      raise ValueError(f'Expected {value} to be compatible with {dtype_hint}.')
    dtype = dtype_hint
  else:
    if isinstance(value, (np.generic, np.ndarray)):
      dtype = value.dtype.type
      # If the value has a dtype of `np.bytes_` or `np.object_`, the serialized
      # dtype should still be a `np.str_`.
      if np.issubdtype(dtype, np.bytes_) or np.issubdtype(dtype, np.object_):
        dtype = np.str_
    else:
      dtype = dtype_utils.infer_dtype(value)

  # Normalize to a numpy value.
  if dtype is not np.str_:
    value = np.asarray(value, dtype)
  else:
    raise NotImplementedError(f'Unexpected dtype found: {dtype}.')

  dtype_pb = dtype_utils.to_proto(dtype)
  shape_pb = array_shape.to_proto(value.shape)
  content = value.tobytes()

  return array_pb2.Array(dtype=dtype_pb, shape=shape_pb, content=content)


def is_compatible_dtype(value: Array, dtype: type[np.generic]) -> bool:
  """Returns `True` if `value` is compatible with `dtype`, otherwise `False`.

  This functions checks that the `value` has the same scalar kind (e.g. integer,
  floating) and has a compatible size (e.g. 32-bits, 16-bits) as `dtype` .

  See https://numpy.org/doc/stable/reference/arrays.scalars.html for more
  information.

  Args:
    value: The value to check.
    dtype: The dtype to check against.
  """
  if isinstance(value, (np.generic, np.ndarray)):
    value_dtype = value.dtype.type
  else:
    value_dtype = type(value)
  if value_dtype is dtype:
    return True

  # Check dtype kind.
  if np.issubdtype(value_dtype, np.bool_):
    # Skip checking dtype size, `np.bool_` does not have a size.
    return dtype is np.bool_
  elif np.issubdtype(value_dtype, np.integer):
    if not np.issubdtype(dtype, np.integer):
      return False
  elif np.issubdtype(value_dtype, np.floating):
    if not np.issubdtype(dtype, np.floating):
      return False
  elif np.issubdtype(value_dtype, np.complexfloating):
    if not np.issubdtype(dtype, np.complexfloating):
      return False
  elif np.issubdtype(value_dtype, np.character) or np.issubdtype(
      value_dtype, np.object_
  ):
    # Skip checking dtype size, `np.str_`, `np.bytes_`, and `np.object_`
    # (null-terminated bytes) have a variable length.
    return dtype is np.str_
  else:
    return False

  # Check dtype size.
  if isinstance(value, (np.generic, np.ndarray)):
    # `np.can_cast` does not support Python scalars and does not does not apply
    # value-based logic to `np.ndarray` or numpy scalars. See
    # https://numpy.org/doc/stable/reference/generated/numpy.can_cast.html for
    # more information.
    return np.can_cast(value.dtype, dtype)
  elif isinstance(value, (int, float, complex)):
    return dtype_utils.can_cast(value, dtype)
  else:
    return False


def is_compatible_shape(value: Array, shape: array_shape.ArrayShape) -> bool:
  """Returns `True` if `value` is compatible with `shape`, otherwise `False`.

  Args:
    value: The value to check.
    shape: The `federated_language.types.ArrayShape` to check against.
  """
  if isinstance(value, np.ndarray):
    return array_shape.is_compatible_with(value.shape, shape)
  else:
    return array_shape.is_shape_scalar(shape)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# limitations under the License.
"""Helpers for creating larger structures out of computing building blocks."""

from typing import Optional

from federated_language.compiler import building_blocks
from federated_language.computation import computation_wrapper
from federated_language.context_stack import context_stack_impl
from federated_language.federated_context import federated_computation_context
from federated_language.federated_context import value_impl
from federated_language.types import computation_types
from federated_language.types import type_conversions


def zero_or_one_arg_fn_to_building_block(
    fn,
    parameter_name: Optional[str],
    parameter_type: Optional[computation_types.Type],
    context_stack: context_stack_impl.ContextStack,
    suggested_name: Optional[str] = None,
) -> tuple[
    building_blocks.ComputationBuildingBlock, computation_types.FunctionType
]:
  """Converts a zero- or one-argument `fn` into a computation building block.

  Args:
    fn: A function with 0 or 1 arguments that contains orchestration logic,
      i.e., that expects zero or one `values_base.Value` and returns a result
      convertible to the same.
    parameter_name: The name of the parameter, or `None` if there is't any.
    parameter_type: The `federated_language.Type` of the parameter, or `None` if
      there's none.
    context_stack: The context stack to use.
    suggested_name: The optional suggested name to use for the federated context
      that will be used to serialize this function's body (ideally the name of
      the underlying Python function). It might be modified to avoid conflicts.

  Returns:
    A tuple of `(building_blocks.ComputationBuildingBlock,
    computation_types.Type)`, where the first element contains the logic from
    `fn`, and the second element contains potentially annotated type information
    for the result of `fn`.

  Raises:
    ValueError: if `fn` is incompatible with `parameter_type`.
  """
  if isinstance(
      context_stack.current,
      federated_computation_context.FederatedComputationContext,
  ):
    parent_context = context_stack.current
  else:
    parent_context = None
  context = federated_computation_context.FederatedComputationContext(
      context_stack, suggested_name=suggested_name, parent=parent_context
  )
  if parameter_name is not None:
    parameter_name = '{}_{}'.format(context.name, str(parameter_name))
  with context_stack.install(context):
    if parameter_type is not None:
      result = fn(
          value_impl.Value(
              building_blocks.Reference(parameter_name, parameter_type),
          )
      )
    else:
      result = fn()
    if result is None:
      raise computation_wrapper.ComputationReturnedNoneError(fn)
    annotated_result_type = type_conversions.infer_type(result)
    result = value_impl.to_value(result, type_spec=annotated_result_type)
    result_comp = result.comp
    symbols_bound_in_context = context_stack.current.symbol_bindings
    if symbols_bound_in_context:
      result_comp = building_blocks.Block(
          local_symbols=symbols_bound_in_context, result=result_comp
      )
    annotated_type = computation_types.FunctionType(
        parameter_type, annotated_result_type
    )
    return (
        building_blocks.Lambda(parameter_name, parameter_type, result_comp),
        annotated_type,
    )



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for working with shapes.

The shape of an `Array` may be one of the following:

*   Fully-defined: Has a known number of dimensions and a known size for each
    dimension (e.g. (2, 3)).
*   Partially-defined: Has a known number of dimensions, and an unknown size for
    one or more dimension (e.g. (2, None)).
*   Unknown: Has an unknown number of dimensions (e.g. None).
*   Scalar: Has no dimensions (e.g. ()).
"""

from collections.abc import Sequence
import functools
import operator
from typing import Optional, Union

from federated_language.proto import array_pb2
from federated_language.proto import data_type_pb2  # pylint: disable=unused-import  # b/330931277


_EmptyTuple = tuple[()]
_ArrayShapeLike = Union[Sequence[Optional[int]], None, _EmptyTuple]

# ArrayShape is the Python representation of the `ArrayShape` protobuf, and is
# the shape of an `Array`.
ArrayShape = Union[tuple[Optional[int], ...], None, _EmptyTuple]


def from_proto(shape_pb: array_pb2.ArrayShape) -> ArrayShape:
  """Returns a `federated_language.types.ArrayShape` for the `shape_pb`."""
  if shape_pb.unknown_rank:
    return None
  else:
    return tuple(d if d >= 0 else None for d in shape_pb.dim)


def to_proto(shape: ArrayShape) -> array_pb2.ArrayShape:
  """Returns an `ArrayShape` for the `shape`."""
  if shape is not None:
    dims = [d if d is not None else -1 for d in shape]
    return array_pb2.ArrayShape(dim=dims)
  else:
    return array_pb2.ArrayShape(unknown_rank=True)


def is_shape_fully_defined(shape: ArrayShape) -> bool:
  """Returns `True` if `shape` is fully defined, False otherwise.

  Args:
    shape: A `federated_language.types.ArrayShape`.
  """
  return shape is not None and all(dim is not None for dim in shape)


def is_shape_scalar(shape: ArrayShape) -> bool:
  """Returns `True` if `shape` is scalar, False otherwise.

  Args:
    shape: A `federated_language.types.ArrayShape`.
  """
  return shape is not None and not shape


def is_compatible_with(target: ArrayShape, other: ArrayShape) -> bool:
  """Returns `True` if `target` is compatible with `other`, otherwise `False`.

  Two shapes are compatible if there exists a fully-defined shape that both
  shapes can represent. For example:

  * `None` is compatible with all shapes.
  * `(None, None)` is compatible with all two-dimensional shapes, and also
    `None`.
  * `(2, None)` is compatible with all two-dimensional shapes with size 2 in the
    0th dimension, and also `(None, None)` and `None`.
  * `(2, 3) is compatible with itself, and also `(32, None)`, `(None, 3]),
    `(None, None)`, and `None`.

  Args:
    target: A `federated_language.types.ArrayShape`.
    other: Another `federated_language.types.ArrayShape`.
  """
  if target is None or other is None:
    return True

  if len(target) != len(other):
    return False

  return all(x is None or y is None or x == y for x, y in zip(target, other))


def num_elements_in_shape(shape: ArrayShape) -> Optional[int]:
  """Returns the number of elements in `shape`, or `None` if not fully defined.

  Args:
    shape: A `federated_language.types.ArrayShape`.
  """
  if is_shape_fully_defined(shape):
    return functools.reduce(operator.mul, shape, 1)
  else:
    return None



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Libraries for extending the Federated Language core library."""

# pylint: disable=g-importing-member
from federated_language.common_libs.tracing import propagate_trace_context_task_factory
from federated_language.common_libs.tracing import set_tracing_providers
from federated_language.common_libs.tracing import span
from federated_language.common_libs.tracing import trace
from federated_language.common_libs.tracing import TraceResult
from federated_language.common_libs.tracing import TracingProvider
from federated_language.common_libs.tracing import wrap_coroutine_in_current_trace_context
from federated_language.common_libs.tracing import wrap_rpc_in_trace_context
from federated_language.compiler.building_block_analysis import is_called_aggregation
from federated_language.compiler.building_block_analysis import is_identity_function
from federated_language.compiler.building_block_factory import create_federated_aggregate
from federated_language.compiler.building_block_factory import create_federated_apply
from federated_language.compiler.building_block_factory import create_federated_broadcast
from federated_language.compiler.building_block_factory import create_federated_map
from federated_language.compiler.building_block_factory import create_federated_map_all_equal
from federated_language.compiler.building_block_factory import create_federated_map_or_apply
from federated_language.compiler.building_block_factory import create_federated_mean
from federated_language.compiler.building_block_factory import create_federated_secure_sum
from federated_language.compiler.building_block_factory import create_federated_secure_sum_bitwidth
from federated_language.compiler.building_block_factory import create_federated_select
from federated_language.compiler.building_block_factory import create_federated_sum
from federated_language.compiler.building_block_factory import create_federated_unzip
from federated_language.compiler.building_block_factory import create_federated_value
from federated_language.compiler.building_block_factory import create_federated_zip
from federated_language.compiler.building_block_factory import create_identity
from federated_language.compiler.building_block_factory import create_named_tuple
from federated_language.compiler.building_block_factory import create_sequence_map
from federated_language.compiler.building_block_factory import Path
from federated_language.compiler.building_block_factory import select_output_from_lambda
from federated_language.compiler.building_block_factory import unique_name_generator
from federated_language.compiler.building_blocks import Block
from federated_language.compiler.building_blocks import Call
from federated_language.compiler.building_blocks import CompiledComputation
from federated_language.compiler.building_blocks import ComputationBuildingBlock
from federated_language.compiler.building_blocks import Data
from federated_language.compiler.building_blocks import Intrinsic
from federated_language.compiler.building_blocks import Lambda
from federated_language.compiler.building_blocks import Literal
from federated_language.compiler.building_blocks import Placement
from federated_language.compiler.building_blocks import Reference
from federated_language.compiler.building_blocks import Selection
from federated_language.compiler.building_blocks import Struct
from federated_language.compiler.building_blocks import UnexpectedBlockError
from federated_language.compiler.intrinsic_defs import AggregationKind
from federated_language.compiler.intrinsic_defs import FEDERATED_AGGREGATE
from federated_language.compiler.intrinsic_defs import FEDERATED_APPLY
from federated_language.compiler.intrinsic_defs import FEDERATED_BROADCAST
from federated_language.compiler.intrinsic_defs import FEDERATED_EVAL_AT_CLIENTS
from federated_language.compiler.intrinsic_defs import FEDERATED_EVAL_AT_SERVER
from federated_language.compiler.intrinsic_defs import FEDERATED_MAP
from federated_language.compiler.intrinsic_defs import FEDERATED_MAP_ALL_EQUAL
from federated_language.compiler.intrinsic_defs import FEDERATED_MAX
from federated_language.compiler.intrinsic_defs import FEDERATED_MEAN
from federated_language.compiler.intrinsic_defs import FEDERATED_MIN
from federated_language.compiler.intrinsic_defs import FEDERATED_SECURE_SELECT
from federated_language.compiler.intrinsic_defs import FEDERATED_SECURE_SUM
from federated_language.compiler.intrinsic_defs import FEDERATED_SECURE_SUM_BITWIDTH
from federated_language.compiler.intrinsic_defs import FEDERATED_SELECT
from federated_language.compiler.intrinsic_defs import FEDERATED_SUM
from federated_language.compiler.intrinsic_defs import FEDERATED_VALUE_AT_CLIENTS
from federated_language.compiler.intrinsic_defs import FEDERATED_VALUE_AT_SERVER
from federated_language.compiler.intrinsic_defs import FEDERATED_WEIGHTED_MEAN
from federated_language.compiler.intrinsic_defs import FEDERATED_ZIP_AT_CLIENTS
from federated_language.compiler.intrinsic_defs import FEDERATED_ZIP_AT_SERVER
from federated_language.compiler.intrinsic_defs import GENERIC_DIVIDE
from federated_language.compiler.intrinsic_defs import GENERIC_MULTIPLY
from federated_language.compiler.intrinsic_defs import GENERIC_PLUS
from federated_language.compiler.intrinsic_defs import get_aggregation_intrinsics
from federated_language.compiler.intrinsic_defs import get_broadcast_intrinsics
from federated_language.compiler.intrinsic_defs import IntrinsicDef
from federated_language.compiler.intrinsic_defs import SEQUENCE_MAP
from federated_language.compiler.transformation_utils import BoundVariableTracker
from federated_language.compiler.transformation_utils import get_map_of_unbound_references
from federated_language.compiler.transformation_utils import SymbolTree
from federated_language.compiler.transformation_utils import transform_postorder
from federated_language.compiler.transformation_utils import transform_postorder_with_symbol_bindings
from federated_language.compiler.transformation_utils import transform_preorder
from federated_language.compiler.transformation_utils import TransformReturnType
from federated_language.compiler.tree_analysis import check_aggregate_not_dependent_on_aggregate
from federated_language.compiler.tree_analysis import check_broadcast_not_dependent_on_aggregate
from federated_language.compiler.tree_analysis import check_contains_no_new_unbound_references
from federated_language.compiler.tree_analysis import check_contains_no_unbound_references
from federated_language.compiler.tree_analysis import check_contains_only_reducible_intrinsics
from federated_language.compiler.tree_analysis import check_has_single_placement
from federated_language.compiler.tree_analysis import check_has_unique_names
from federated_language.compiler.tree_analysis import contains_called_intrinsic
from federated_language.compiler.tree_analysis import contains_no_unbound_references
from federated_language.compiler.tree_analysis import count as computation_count
from federated_language.compiler.tree_analysis import find_aggregations_in_tree
from federated_language.compiler.tree_analysis import NonuniqueNameError
from federated_language.compiler.tree_analysis import visit_postorder
from federated_language.compiler.tree_analysis import visit_preorder
from federated_language.computation.computation_analysis import computation_contains
from federated_language.computation.computation_analysis import computation_contains_secure_aggregation
from federated_language.computation.computation_analysis import computation_contains_unsecure_aggregation
from federated_language.computation.computation_base import Computation
from federated_language.computation.computation_impl import ConcreteComputation
from federated_language.computation.computation_wrapper import ComputationReturnedNoneError
from federated_language.computation.computation_wrapper import ComputationWrapper
from federated_language.computation.function_utils import pack_args_into_struct
from federated_language.computation.function_utils import unpack_arg
from federated_language.computation.function_utils import unpack_args_from_struct
from federated_language.computation.function_utils import wrap_as_zero_or_one_arg_callable
from federated_language.computation.polymorphic_computation import PolymorphicComputation
from federated_language.context_stack.context import AsyncContext
from federated_language.context_stack.context import ContextError
from federated_language.context_stack.context import SyncContext
from federated_language.context_stack.context_stack_impl import ContextStack
from federated_language.context_stack.context_stack_impl import get_context_stack
from federated_language.context_stack.context_stack_impl import set_default_context
from federated_language.context_stack.context_stack_impl import set_no_default_context
from federated_language.context_stack.context_stack_test_utils import with_context
from federated_language.context_stack.runtime_error_context import RuntimeErrorContext
from federated_language.context_stack.symbol_binding_context import SymbolBindingContext
from federated_language.executor.async_execution_context import AsyncExecutionContext
from federated_language.executor.cardinalities_utils import infer_cardinalities
from federated_language.executor.executor_base import Executor
from federated_language.executor.executor_base import RetryableError
from federated_language.executor.executor_factory import CardinalitiesType
from federated_language.executor.executor_factory import ExecutorFactory
from federated_language.executor.executor_value_base import ExecutorValue
from federated_language.executor.sync_execution_context import SyncExecutionContext
from federated_language.federated_context.federated_computation_context import FederatedComputationContext
from federated_language.federated_context.value_utils import ensure_federated_value
from federated_language.types.computation_types import TypeNotAssignableError
from federated_language.types.computation_types import TypesNotEquivalentError
from federated_language.types.computation_types import UnexpectedTypeError
from federated_language.types.placements import PlacementLiteral
from federated_language.types.placements import uri_to_placement_literal
from federated_language.types.type_analysis import check_concrete_instance_of
from federated_language.types.type_analysis import check_is_structure_of_integers
from federated_language.types.type_analysis import contains as type_contains
from federated_language.types.type_analysis import contains_federated_types
from federated_language.types.type_analysis import contains_only as type_contains_only
from federated_language.types.type_analysis import contains_tensor_types
from federated_language.types.type_analysis import count as type_count
from federated_language.types.type_analysis import is_generic_op_compatible_type
from federated_language.types.type_analysis import is_single_integer_or_matches_structure
from federated_language.types.type_analysis import is_structure_of_floats
from federated_language.types.type_analysis import is_structure_of_integers
from federated_language.types.type_analysis import is_structure_of_tensors
from federated_language.types.type_analysis import preorder_types
from federated_language.types.type_conversions import infer_type
from federated_language.types.type_conversions import to_structure_with_type
from federated_language.types.type_conversions import type_to_py_container
from federated_language.types.type_transformations import transform_type_postorder
# pylint: enable=g-importing-member



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A library of static analysis functions for computation types."""

import collections
from collections.abc import Callable
from typing import Optional

from federated_language.types import array_shape
from federated_language.types import computation_types
from federated_language.types import placements
import ml_dtypes
import numpy as np

_TypePredicate = Callable[[computation_types.Type], bool]


def preorder_types(type_signature: computation_types.Type):
  """Yields each type in `type_signature` in a preorder fashion."""
  yield type_signature
  for child in type_signature.children():
    yield from preorder_types(child)


def count(
    type_signature: computation_types.Type, predicate: _TypePredicate
) -> int:
  """Returns the number of types in `type_signature` matching `predicate`.

  Args:
    type_signature: A tree of `computation_type.Type`s to count.
    predicate: A Python function that takes a type as a parameter and returns a
      boolean value.
  """
  one_or_zero = lambda t: 1 if predicate(t) else 0
  return sum(map(one_or_zero, preorder_types(type_signature)))


def contains(
    type_signature: computation_types.Type, predicate: _TypePredicate
) -> bool:
  """Checks if `type_signature` contains any types that pass `predicate`."""
  for t in preorder_types(type_signature):
    if predicate(t):
      return True
  return False


def contains_federated_types(type_signature):
  """Returns whether or not `type_signature` contains a federated type."""
  return contains(
      type_signature, lambda t: isinstance(t, computation_types.FederatedType)
  )


def contains_tensor_types(type_signature):
  """Returns whether or not `type_signature` contains a tensor type."""
  return contains(
      type_signature, lambda t: isinstance(t, computation_types.TensorType)
  )


def contains_only(
    type_signature: computation_types.Type,
    predicate: _TypePredicate,
) -> bool:
  """Checks if `type_signature` contains only types that pass `predicate`."""
  return not contains(type_signature, lambda t: not predicate(t))


def is_structure_of_tensors(type_spec: computation_types.Type) -> bool:
  def _predicate(type_spec: computation_types.Type) -> bool:
    return isinstance(
        type_spec,
        (
            computation_types.StructType,
            computation_types.TensorType,
        ),
    )

  return contains_only(type_spec, _predicate)


def is_generic_op_compatible_type(type_spec):
  """Checks `type_spec` against an explicit list of generic operators."""
  if type_spec is None:
    return False

  def _predicate(type_spec: computation_types.Type) -> bool:
    return isinstance(
        type_spec,
        (
            computation_types.TensorType,
            computation_types.StructType,
        ),
    )

  return contains_only(type_spec, _predicate)


class SumIncompatibleError(TypeError):

  def __init__(self, type_spec, type_spec_context, reason):
    message = (
        'Expected a type which is compatible with the sum operator, found\n'
        f'{type_spec_context}\nwhich contains\n{type_spec}\nwhich is not '
        f'sum-compatible because {reason}.'
    )
    super().__init__(message)


def check_is_sum_compatible(
    type_spec: computation_types.Type,
    type_spec_context: Optional[computation_types.Type] = None,
):
  """Determines if `type_spec` is a type that can be added to itself.

  Types that are sum-compatible are composed of scalars of numeric types,
  possibly packaged into nested named tuples, and possibly federated. Types
  that are sum-incompatible include sequences, functions, abstract types,
  and placements.

  Args:
    type_spec: A `computation_types.Type`.
    type_spec_context: An optional parent type to include in the error message.

  Raises:
     SumIncompatibleError: if `type_spec` is not sum-compatible.
  """
  if type_spec_context is None:
    type_spec_context = type_spec
  if isinstance(type_spec, computation_types.TensorType):
    if not (
        np.issubdtype(type_spec.dtype, np.number)
        or type_spec.dtype == ml_dtypes.bfloat16
    ):
      raise SumIncompatibleError(
          type_spec, type_spec_context, f'{type_spec.dtype} is not numeric'
      )
    if not array_shape.is_shape_fully_defined(type_spec.shape):
      raise SumIncompatibleError(
          type_spec,
          type_spec_context,
          f'{type_spec.shape} is not fully defined',
      )
  elif isinstance(type_spec, computation_types.StructType):
    for _, element_type in type_spec.items():
      check_is_sum_compatible(element_type, type_spec_context)
  elif isinstance(type_spec, computation_types.FederatedType):
    check_is_sum_compatible(type_spec.member, type_spec_context)
  else:
    raise SumIncompatibleError(
        type_spec,
        type_spec_context,
        'only structures of tensors (possibly federated) may be summed',
    )


def is_structure_of_floats(type_spec: computation_types.Type) -> bool:
  """Determines if `type_spec` is a structure of floats.

  Note that an empty `computation_types.StructType` will return `True`, as it
  does not contain any non-floating types.

  Args:
    type_spec: A `computation_types.Type`.

  Returns:
    `True` iff `type_spec` is a structure of floats, otherwise `False`.
  """
  if isinstance(type_spec, computation_types.TensorType):
    return np.issubdtype(type_spec.dtype, np.floating)
  elif isinstance(type_spec, computation_types.StructType):
    return all(is_structure_of_floats(v) for _, v in type_spec.items())
  elif isinstance(type_spec, computation_types.FederatedType):
    return is_structure_of_floats(type_spec.member)
  else:
    return False


def is_structure_of_integers(type_spec: computation_types.Type) -> bool:
  """Determines if `type_spec` is a structure of integers.

  Note that an empty `computation_types.StructType` will return `True`, as it
  does not contain any non-integer types.

  Args:
    type_spec: A `computation_types.Type`.

  Returns:
    `True` iff `type_spec` is a structure of integers, otherwise `False`.
  """
  if isinstance(type_spec, computation_types.TensorType):
    return np.issubdtype(type_spec.dtype, np.integer)
  elif isinstance(type_spec, computation_types.StructType):
    return all(is_structure_of_integers(v) for _, v in type_spec.items())
  elif isinstance(type_spec, computation_types.FederatedType):
    return is_structure_of_integers(type_spec.member)
  else:
    return False


def check_is_structure_of_integers(type_spec):
  if not is_structure_of_integers(type_spec):
    raise TypeError(
        'Expected a type which is structure of integers, found {}.'.format(
            type_spec
        )
    )


def is_single_integer_or_matches_structure(
    type_sig: computation_types.Type, shape_type: computation_types.Type
) -> bool:
  """If `type_sig` is an integer or integer structure matching `shape_type`."""
  if isinstance(type_sig, computation_types.TensorType):
    # This condition applies to both `shape_type` being a tensor or structure,
    # as the same integer bitwidth can be used for all values in the structure.
    return (
        np.issubdtype(type_sig.dtype, np.integer)
        and array_shape.num_elements_in_shape(type_sig.shape) == 1
    )
  elif isinstance(shape_type, computation_types.StructType) and isinstance(
      type_sig, computation_types.StructType
  ):
    bitwidth_name_and_types = list(type_sig.items())
    shape_name_and_types = list(shape_type.items())
    if len(type_sig) != len(shape_name_and_types):
      return False
    for (inner_name, type_sig), (inner_shape_name, inner_shape_type) in zip(
        bitwidth_name_and_types, shape_name_and_types
    ):
      if inner_name != inner_shape_name:
        return False
      if not is_single_integer_or_matches_structure(type_sig, inner_shape_type):
        return False
    return True
  else:
    return False


def check_federated_type(
    type_spec: computation_types.FederatedType,
    member: Optional[computation_types.Type] = None,
    placement: Optional[placements.PlacementLiteral] = None,
    all_equal: Optional[bool] = None,
):
  """Checks that `type_spec` is a federated type with the given parameters.

  Args:
    type_spec: The `federated_language.FederatedType` to check.
    member: The expected member type, or `None` if unspecified.
    placement: The desired placement, or `None` if unspecified.
    all_equal: The desired result of accessing the property
      `federated_language.FederatedType.all_equal` of `type_spec`, or `None` if
      left unspecified.

  Raises:
    TypeError: if `type_spec` is not a federated type of the given kind.
  """
  if member is not None:
    member.check_assignable_from(type_spec.member)
  if placement is not None:
    if type_spec.placement is not placement:
      raise TypeError(
          'Expected federated type placed at {}, got one placed at {}.'.format(
              placement, type_spec.placement
          )
      )
  if all_equal is not None:
    if type_spec.all_equal != all_equal:
      raise TypeError(
          'Expected federated type with all_equal {}, got one with {}.'.format(
              all_equal, type_spec.all_equal
          )
      )


def is_average_compatible(type_spec: computation_types.Type) -> bool:
  """Determines if `type_spec` can be averaged.

  Types that are average-compatible are composed of numeric tensor types,
  either floating-point or complex, possibly packaged into nested named tuples,
  and possibly federated.

  Args:
    type_spec: a `computation_types.Type`.

  Returns:
    `True` iff `type_spec` is average-compatible, `False` otherwise.
  """
  if isinstance(type_spec, computation_types.TensorType):
    return np.issubdtype(type_spec, np.inexact)
  elif isinstance(type_spec, computation_types.StructType):
    return all(is_average_compatible(v) for _, v in type_spec.items())
  elif isinstance(type_spec, computation_types.FederatedType):
    return is_average_compatible(type_spec.member)
  else:
    return False


def is_min_max_compatible(type_spec: computation_types.Type) -> bool:
  """Determines if `type_spec` is min/max compatible.

  Types that are min/max-compatible are composed of integer or floating tensor
  types, possibly packaged into nested tuples and possibly federated.

  Args:
    type_spec: a `computation_types.Type`.

  Returns:
    `True` iff `type_spec` is min/max compatible, `False` otherwise.
  """
  if isinstance(type_spec, computation_types.TensorType):
    return np.issubdtype(type_spec.dtype, np.integer) or np.issubdtype(
        type_spec.dtype, np.floating
    )
  elif isinstance(type_spec, computation_types.StructType):
    return all(is_min_max_compatible(v) for _, v in type_spec.items())
  elif isinstance(type_spec, computation_types.FederatedType):
    return is_min_max_compatible(type_spec.member)
  else:
    return False


class NotConcreteTypeError(TypeError):

  def __init__(self, full_type, found_abstract):
    message = (
        'Expected concrete type containing no abstract types, but '
        f'found abstract type {found_abstract} in {full_type}.'
    )
    super().__init__(message)


class MismatchedConcreteTypesError(TypeError):
  """Raised when there is a mismatch between two types."""

  def __init__(
      self,
      full_concrete,
      full_generic,
      abstract_label,
      first_concrete,
      second_concrete,
  ):
    message = (
        f'Expected concrete type {full_concrete} to be a valid substitution '
        f'for generic type {full_generic}, but abstract type {abstract_label} '
        f'had substitutions {first_concrete} and {second_concrete}, which are '
        'not equivalent.'
    )
    super().__init__(message)


class UnassignableConcreteTypesError(TypeError):
  """Raised when one type can not be assigned to another type."""

  def __init__(
      self,
      full_concrete,
      full_generic,
      abstract_label,
      definition,
      not_assignable_from,
  ):
    message = (
        f'Expected concrete type {full_concrete} to be a valid substitution '
        f'for generic type {full_generic}, but abstract type {abstract_label} '
        f'was defined as {definition}, and later used as {not_assignable_from} '
        'which cannot be assigned from the former.'
    )
    super().__init__(message)


class MismatchedStructureError(TypeError):
  """Raised when there is a mismatch between the structures of two types."""

  def __init__(
      self,
      full_concrete,
      full_generic,
      concrete_member,
      generic_member,
      mismatch,
  ):
    message = (
        f'Expected concrete type {full_concrete} to be a valid substitution '
        f'for generic type {full_generic}, but their structures do not match: '
        f'{concrete_member} differs in {mismatch} from {generic_member}.'
    )
    super().__init__(message)


class MissingDefiningUsageError(TypeError):

  def __init__(self, generic_type, label_name):
    message = (
        f'Missing defining use of abstract type {label_name} in type '
        f'{generic_type}. See `check_concrete_instance_of` documentation for '
        'details on what counts as a defining use.'
    )
    super().__init__(message)


def check_concrete_instance_of(
    concrete_type: computation_types.Type, generic_type: computation_types.Type
):
  """Checks whether `concrete_type` is a valid substitution of `generic_type`.

  This function determines whether `generic_type`'s type parameters can be
  substituted such that it is equivalent to `concrete type`.

  Note that passing through argument-position of function type swaps the
  variance of abstract types. Argument-position types can be assigned *from*
  other instances of the same type, but are not equivalent to it.

  Due to this variance issue, only abstract types must include at least one
  "defining" usage. "Defining" uses are those which are encased in function
  parameter position an odd number of times. These usages must all be
  equivalent. Non-defining usages need not compare equal but must be assignable
  *from* defining usages.

  Args:
    concrete_type: A type containing no `computation_types.AbstractType`s to
      check against `generic_type`'s shape.
    generic_type: A type which may contain `computation_types.AbstractType`s.

  Raises:
    TypeError: If `concrete_type` is not a valid substitution of `generic_type`.
  """
  for t in preorder_types(concrete_type):
    if isinstance(t, computation_types.AbstractType):
      raise NotConcreteTypeError(concrete_type, t)

  type_bindings = {}
  non_defining_usages = collections.defaultdict(list)

  def _check_helper(
      generic_type_member: computation_types.Type,
      concrete_type_member: computation_types.Type,
      defining: bool,
  ):
    """Recursive helper function."""

    def _raise_structural(mismatch):
      raise MismatchedStructureError(
          concrete_type,
          generic_type,
          concrete_type_member,
          generic_type_member,
          mismatch,
      )

    def _both_are(predicate):
      if predicate(generic_type_member):
        if predicate(concrete_type_member):
          return True
        else:
          _raise_structural('kind')
      else:
        return False

    if isinstance(generic_type_member, computation_types.AbstractType):
      label = str(generic_type_member.label)
      if not defining:
        non_defining_usages[label].append(concrete_type_member)
      else:
        bound_type = type_bindings.get(label)
        if bound_type is not None:
          if not concrete_type_member.is_equivalent_to(bound_type):
            raise MismatchedConcreteTypesError(
                concrete_type,
                generic_type,
                label,
                bound_type,
                concrete_type_member,
            )
        else:
          type_bindings[label] = concrete_type_member
    elif _both_are(lambda t: isinstance(t, computation_types.TensorType)):
      if generic_type_member != concrete_type_member:
        _raise_structural('tensor types')
    elif _both_are(lambda t: isinstance(t, computation_types.PlacementType)):
      if generic_type_member != concrete_type_member:
        _raise_structural('placements')
    elif _both_are(lambda t: isinstance(t, computation_types.StructType)):
      generic_elements = list(generic_type_member.items())  # pytype: disable=attribute-error
      concrete_elements = list(concrete_type_member.items())  # pytype: disable=attribute-error
      if len(generic_elements) != len(concrete_elements):
        _raise_structural('length')
      for generic_element, concrete_element in zip(
          generic_elements, concrete_elements
      ):
        if generic_element[0] != concrete_element[0]:
          _raise_structural('element names')
        _check_helper(generic_element[1], concrete_element[1], defining)
    elif _both_are(lambda t: isinstance(t, computation_types.SequenceType)):
      _check_helper(
          generic_type_member.element,  # pytype: disable=attribute-error
          concrete_type_member.element,  # pytype: disable=attribute-error
          defining,
      )
    elif _both_are(lambda t: isinstance(t, computation_types.FunctionType)):
      if generic_type_member.parameter is None:  # pytype: disable=attribute-error
        if concrete_type_member.parameter is not None:  # pytype: disable=attribute-error
          _raise_structural('parameter')
      else:
        _check_helper(
            generic_type_member.parameter,  # pytype: disable=attribute-error
            concrete_type_member.parameter,  # pytype: disable=attribute-error
            not defining,
        )
      _check_helper(
          generic_type_member.result,  # pytype: disable=attribute-error
          concrete_type_member.result,  # pytype: disable=attribute-error
          defining,
      )
    elif _both_are(lambda t: isinstance(t, computation_types.FederatedType)):
      if generic_type_member.placement != concrete_type_member.placement:  # pytype: disable=attribute-error
        _raise_structural('placement')
      if generic_type_member.all_equal != concrete_type_member.all_equal:  # pytype: disable=attribute-error
        _raise_structural('all equal')
      _check_helper(
          generic_type_member.member,  # pytype: disable=attribute-error
          concrete_type_member.member,  # pytype: disable=attribute-error
          defining,
      )
    else:
      raise TypeError(f'Unexpected type kind {generic_type}.')

  _check_helper(generic_type, concrete_type, False)

  for label, usages in non_defining_usages.items():
    bound_type = type_bindings.get(label)
    if bound_type is None:
      if len(usages) == 1:
        # Single-use abstract types can't be wrong.
        # NOTE: we could also add an exception here for cases where every usage
        # is equivalent to the first usage. However, that's not currently
        # needed since the only intrinsic that doesn't have a defining use is
        # GENERIC_ZERO, which has only a single-use type parameter.
        pass
      else:
        raise MissingDefiningUsageError(generic_type, label)
    else:
      for usage in usages:
        if not usage.is_assignable_from(bound_type):
          raise UnassignableConcreteTypesError(
              concrete_type, generic_type, label, bound_type, usage
          )



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Representation of values inside a federated computation."""

import abc
from collections.abc import Hashable, Mapping
import itertools
import typing
from typing import Optional, Union

import attrs
from federated_language.common_libs import py_typecheck
from federated_language.common_libs import structure
from federated_language.compiler import array
from federated_language.compiler import building_block_factory
from federated_language.compiler import building_blocks
from federated_language.computation import computation_impl
from federated_language.computation import function_utils
from federated_language.computation import polymorphic_computation
from federated_language.context_stack import context
from federated_language.context_stack import context_stack_impl
from federated_language.context_stack import symbol_binding_context
from federated_language.types import computation_types
from federated_language.types import placements
from federated_language.types import type_conversions
from federated_language.types import typed_object


def _unfederated(type_signature):
  if isinstance(type_signature, computation_types.FederatedType):
    return type_signature.member
  return type_signature


def _is_federated_struct(type_spec: computation_types.Type) -> bool:
  return isinstance(type_spec, computation_types.FederatedType) and isinstance(
      type_spec.member, computation_types.StructType
  )


def _check_struct_or_federated_struct(
    vimpl: 'Value',
    attribute: str,
):
  if not isinstance(
      vimpl.type_signature, computation_types.StructType
  ) and not _is_federated_struct(vimpl.type_signature):
    raise AttributeError(
        '`federated_language.Value` of non-structural type'
        f' {vimpl.type_signature} has no attribute {attribute}'
    )


def _bind_computation_to_reference(comp, op: str):
  current_context = context_stack_impl.context_stack.current
  if not isinstance(
      current_context, symbol_binding_context.SymbolBindingContext
  ):
    raise context.ContextError(
        '`federated_language.Value`s should only be used in contexts which can'
        ' bind references, generally a `FederatedComputationContext`.'
        f' Attempted to bind the result of {op} in a context'
        f' {current_context} of type {type(current_context)}.'
    )
  return current_context.bind_computation_to_reference(comp)


class Value(typed_object.TypedObject, abc.ABC):
  """A generic base class for values that appear in computations.

  If the value in this class is of `StructType` or `FederatedType` containing a
  `StructType`, the inner fields can be accessed by name
  (e.g. `y = my_value_impl.y`).
  """

  def __init__(
      self,
      comp: building_blocks.ComputationBuildingBlock,
  ):
    """Constructs a value of the given type.

    Args:
      comp: An instance of building_blocks.ComputationBuildingBlock that
        contains the logic that computes this value.
    """
    super()
    self._comp = comp

  @property
  def type_signature(self):
    return self._comp.type_signature

  @property
  def comp(self) -> building_blocks.ComputationBuildingBlock:
    return self._comp

  def __repr__(self):
    return repr(self._comp)

  def __str__(self):
    return str(self._comp)

  def __dir__(self):
    attributes = ['type_signature', 'comp']
    type_signature = _unfederated(self.type_signature)
    if isinstance(type_signature, computation_types.StructType):
      attributes.extend(dir(type_signature))
    return attributes

  def __getattr__(self, name: str) -> 'Value':
    _check_struct_or_federated_struct(self, name)
    if _is_federated_struct(self.type_signature):
      if name not in self.type_signature.member.fields():  # pytype: disable=attribute-error
        raise AttributeError(
            f"There is no such attribute '{name}' in this federated tuple."
            f" Valid attributes: ({', '.join(dir(self.type_signature.member))})"  # pytype: disable=attribute-error
        )

      return Value(
          building_block_factory.create_federated_getattr_call(self._comp, name)
      )
    if name not in dir(self.type_signature):
      attributes = ', '.join(dir(self.type_signature))
      raise AttributeError(
          f"There is no such attribute '{name}' in this tuple. Valid "
          f'attributes: ({attributes})'
      )
    if isinstance(self._comp, building_blocks.Struct):
      return Value(getattr(self._comp, name))
    return Value(building_blocks.Selection(self._comp, name=name))

  def __bool__(self):
    raise TypeError(
        'Federated computation values do not support boolean operations. '
        'If you were attempting to perform logic on tensors, consider moving '
        'this logic into a `federated_language.tensorflow.computation`.'
    )

  def __len__(self):
    type_signature = _unfederated(self.type_signature)
    if not isinstance(type_signature, computation_types.StructType):
      raise TypeError(
          'Operator len() is only supported for (possibly federated) structure '
          'types, but the object on which it has been invoked is of type {}.'
          .format(self.type_signature)
      )
    return len(type_signature)

  def __getitem__(self, key: Union[int, str, slice]):
    if isinstance(key, str):
      return getattr(self, key)
    if _is_federated_struct(self.type_signature):
      return Value(
          building_block_factory.create_federated_getitem_call(self._comp, key),
      )
    if not isinstance(self.type_signature, computation_types.StructType):
      raise TypeError(
          'Operator getitem() is only supported for structure types, but the '
          'object on which it has been invoked is of type {}.'.format(
              self.type_signature
          )
      )
    elem_length = len(self.type_signature)
    if isinstance(key, int):
      if key < 0 or key >= elem_length:
        raise IndexError(
            'The index of the selected element {} is out of range.'.format(key)
        )
      if isinstance(self._comp, building_blocks.Struct):
        return Value(self._comp[key])
      else:
        return Value(building_blocks.Selection(self._comp, index=key))
    elif isinstance(key, slice):
      index_range = range(*key.indices(elem_length))
      if not index_range:
        raise IndexError(
            'Attempted to slice 0 elements, which is not currently supported.'
        )
      return to_value([self[k] for k in index_range], None)

  def __iter__(self):
    type_signature = _unfederated(self.type_signature)
    if not isinstance(type_signature, computation_types.StructType):
      raise TypeError(
          'Operator iter() is only supported for (possibly federated) '
          'structure types, but the object on which it has been invoked is of '
          f'type {self.type_signature}.'
      )
    for index in range(len(type_signature)):
      yield self[index]

  def __call__(self, *args, **kwargs):
    if not isinstance(self.type_signature, computation_types.FunctionType):
      raise SyntaxError(
          'Function-like invocation is only supported for values of functional '
          'types, but the value being invoked is of type {} that does not '
          'support invocation.'.format(self.type_signature)
      )
    if args or kwargs:
      args = [to_value(x, None) for x in args]
      kwargs = {k: to_value(v, None) for k, v in kwargs.items()}
      arg = function_utils.pack_args(
          self.type_signature.parameter, args, kwargs
      )
      arg = to_value(arg, None).comp
    else:
      arg = None
    call = building_blocks.Call(self._comp, arg)
    ref = _bind_computation_to_reference(
        call, 'calling a `federated_language.Value`'
    )
    return Value(ref)


def _dictlike_items_to_value(items, type_spec, container_type) -> Value:
  elements = []
  for i, (k, v) in enumerate(items):
    element_type = None if type_spec is None else type_spec[i]  # pytype: disable=unsupported-operands
    element_value = to_value(v, element_type)
    elements.append((k, element_value.comp))
  return Value(building_blocks.Struct(elements, container_type))


def to_value(
    arg: object,
    type_spec: Optional[computation_types.Type],
    *,
    parameter_type_hint=None,
    zip_if_needed: bool = False,
) -> Value:
  """Converts the argument into an instance of the abstract class `federated_language.Value`.

  Instances of `federated_language.Value` represent values that appear
  internally in federated computations. This helper function can be used to wrap
  a variety of Python objects as `federated_language.Value` instances to allow
  them to be passed as arguments, used as functions, or otherwise manipulated
  within bodies of federated computations.

  At the moment, the supported types include:

  * Simple constants of `str`, `int`, `float`, and `bool` types, mapped to
    values of a tensor type.

  * Numpy arrays (`np.ndarray` objects), also mapped to tensor type.

  * Dictionaries (`collections.OrderedDict` and unordered `dict`), `list`s,
    `tuple`s, `namedtuple`s, and `Struct`s, all of which are mapped to
    tuple type.

  * Computations (constructed with either the
    `federated_language.tensorflow.computation` or with the
    `federated_language.federated_computation` decorator), typically mapped to
    functions.

  * Placement literals (`federated_language.CLIENTS`,
    `federated_language.SERVER`), mapped to values of the placement type.

  This function is also invoked when attempting to execute a computation.
  All arguments supplied in the invocation are converted into values prior to
  execution. The types of Python objects that can be passed as arguments to
  computations thus matches the types listed here.

  Args:
    arg: An instance of one of the Python types that are convertible to values
      (instances of `federated_language.Value`).
    type_spec: An optional type specifier that allows for disambiguating the
      target type (e.g., when two types can be mapped to the same Python
      representations). If not specified, tried to determine the type of the
      value automatically.
    parameter_type_hint: An optional `federated_language.Type` or value
      convertible to it by `federated_language.to_type()` which specifies an
      argument type to use in the case that `arg` is a
      `polymorphic_computation.PolymorphicComputation`.
    zip_if_needed: If `True`, attempt to coerce the result of `to_value` to
      match `type_spec` by applying `intrinsics.federated_zip` to appropriate
      elements.

  Returns:
    An instance of `federated_language.Value` as described above.

  Raises:
    TypeError: if `arg` is of an unsupported type, or of a type that does not
      match `type_spec`. Raises explicit error message if backend constructs
      are encountered, as backend code should be sealed away from federated
      context.
  """
  # TODO: b/224484886 - Downcasting to all handled types.
  arg = typing.cast(
      Union[
          None,
          Value,
          building_blocks.ComputationBuildingBlock,
          placements.PlacementLiteral,
          computation_impl.ConcreteComputation,
          polymorphic_computation.PolymorphicComputation,
          computation_types.SequenceType,
          structure.Struct,
          py_typecheck.SupportsNamedTuple,
          Mapping[Hashable, object],
          tuple[object, ...],
          list[object],
          array.Array,
      ],
      arg,
  )
  if isinstance(arg, Value):
    result = arg
  elif isinstance(arg, building_blocks.ComputationBuildingBlock):
    result = Value(arg)
  elif isinstance(arg, placements.PlacementLiteral):
    result = Value(building_blocks.Placement(arg))
  elif isinstance(
      arg,
      (
          computation_impl.ConcreteComputation,
          polymorphic_computation.PolymorphicComputation,
      ),
  ):
    if isinstance(arg, polymorphic_computation.PolymorphicComputation):
      if parameter_type_hint is None:
        raise TypeError(
            'Polymorphic computations cannot be converted to'
            ' `federated_language.Value`s without a type hint. Consider'
            ' explicitly specifying the argument types of a computation before'
            ' passing it to a function that requires a'
            ' `federated_language.Value` (such as a intrinsic like'
            ' `federated_map`). If you are a developer and think this'
            ' should be supported, consider providing `parameter_type_hint` as'
            ' an argument to the encompassing `to_value` conversion.'
        )
      parameter_type_hint = computation_types.to_type(parameter_type_hint)
      arg = arg.fn_for_argument_type(parameter_type_hint)
    result = Value(arg.to_compiled_building_block())
  elif isinstance(arg, structure.Struct):
    items = structure.iter_elements(arg)
    result = _dictlike_items_to_value(items, type_spec, None)
  elif isinstance(arg, py_typecheck.SupportsNamedTuple):
    items = arg._asdict().items()
    result = _dictlike_items_to_value(items, type_spec, type(arg))
  elif attrs.has(type(arg)):
    items = attrs.asdict(arg, recurse=False).items()
    result = _dictlike_items_to_value(items, type_spec, type(arg))
  elif isinstance(arg, Mapping):
    result = _dictlike_items_to_value(arg.items(), type_spec, type(arg))
  elif isinstance(arg, (tuple, list)) and not isinstance(
      type_spec, computation_types.SequenceType
  ):
    items = zip(itertools.repeat(None), arg)
    result = _dictlike_items_to_value(items, type_spec, type(arg))
  elif isinstance(arg, typing.get_args(array.Array)):
    if type_spec is None:
      type_spec = type_conversions.infer_type(arg)
    if not isinstance(type_spec, computation_types.TensorType):
      raise ValueError(
          f'Expected a `federated_language.TensorType`, found {type_spec}.'
      )
    literal = building_blocks.Literal(arg, type_spec)
    result = Value(literal)
  else:
    raise TypeError(
        'Expected a Python types that is convertible to a'
        f' `federated_language.Value`, found {type(arg)}. If this is'
        ' backend-specific constructs, it was encountered in a federated'
        ' context and does not support mixing backend-specific and'
        ' federated logic. Please wrap any  backend-specific constructs in a'
        ' computation function.'
    )

  if type_spec is not None and not type_spec.is_assignable_from(
      result.type_signature
  ):
    if zip_if_needed:
      # Returns `None` if such a zip can't be performed.
      zipped_comp = building_block_factory.zip_to_match_type(
          comp_to_zip=result.comp, target_type=type_spec
      )
      if zipped_comp is not None:
        return Value(zipped_comp)
    raise computation_types.TypeNotAssignableError(
        type_spec, result.type_signature
    )
  return result
